import functools
import itertools
from typing import Dict, Optional, Union, Set

from magma.bits import Bits
from magma.circuit import Circuit, CircuitType, CircuitBuilder
from magma.clock import Enable
from magma.conversions import from_bits
from magma.digital import Digital
from magma.primitives.register import AbstractRegister
from magma.ref import DerivedRef
from magma.t import Type, In, Out
from magma.when import (
    BlockBase as WhenBlock,
    no_when,
    find_inferred_latches,
)
from magma.value_utils import make_selector
from magma.wire import wire


_ISWHEN_KEY = "_iswhen_"


def iswhen(inst_or_defn: Union[Circuit, CircuitType]) -> bool:
    if isinstance(inst_or_defn, Circuit):
        return getattr(type(inst_or_defn), _ISWHEN_KEY, False)
    return getattr(inst_or_defn, _ISWHEN_KEY, False)


class InferredLatchError(Exception):
    pass


# NOTE(rsetaluri): This is needed because importing magma.primitives.Memory
# results in a circular import.
@functools.lru_cache(maxsize=1)
def _Memory():
    from magma.primitives.memory import Memory
    return Memory


def _is_memory_port(value: Type) -> bool:
    Memory = _Memory()
    root_ref = value.name.root()
    try:
        value_inst = root_ref.inst
    except AttributeError:
        return False
    if not isinstance(type(value_inst), Memory):
        return False
    if root_ref.name not in ("WDATA", "WADDR", "WE", "RADDR"):
        return False
    return True


def _add_default_drivers_to_memory_ports(
    builder: 'WhenBuilder', latches: Set[Type]
):
    for drivee in builder.output_to_index:
        if drivee in builder.default_drivers:
            continue
        if not _is_memory_port(drivee):
            continue
        if drivee not in latches:
            continue
        T = type(drivee).undirected_t
        zero = from_bits(T, Bits[T.flat_length()](0))
        builder.block.add_default_driver(drivee, zero)
        latches.remove(drivee)


def _get_corresponding_register_default(value: Type) -> Optional[Type]:
    """Used to get the corresponding default for `value` if it is a
    reference to a register input value.

    `value` should either be reg.I or reg.CE:
        * For reg.I, the default will be reg.O (default update is the
          current value).
        * For reg.CE, the default will be False (default to no update).

    """
    root_ref = value.name.root()
    try:
        value_inst = root_ref.inst
    except AttributeError:
        return None
    value_inst_T = type(value_inst)
    if not isinstance(value_inst_T, AbstractRegister):
        return None
    if value is value_inst_T.get_enable(value_inst):
        return type(value)(False)
    sel = make_selector(value)
    return sel.select(value_inst.O)


def _add_default_drivers_to_register_inputs(
    builder: 'WhenBuilder', latches: Set[Type]
):
    """Adds default drivers to register inputs (reg.O is the default value for
    inputs, False is the default for enables).

    The current implementation assumes the magma internally defined register
    primitive is used, which always has only one input and an optional enable
    port. If we encounter a AbstractRegister, we will drive all inputs with
    reg.O and all enables with reg.I (if they have multiple).

    If we wanted to make this interface more extensible, we could have the user
    provide an API for setting up default values.
    """
    for drivee in builder.output_to_index:
        if drivee not in latches:
            continue
        O = _get_corresponding_register_default(drivee)
        if O is None:
            continue
        builder.block.add_default_driver(drivee, O)
        latches.remove(drivee)


class WhenBuilder(CircuitBuilder):
    def __init__(self, block: WhenBlock):
        super().__init__(name=f"When_{id(block)}")
        self._block = block
        self._condition_counter = itertools.count()
        self._driver_counter = itertools.count()
        self._drivee_counter = itertools.count()
        self._input_counter = itertools.count()
        self._output_counter = itertools.count()
        self._input_to_index = {}
        self._output_to_index = {}
        self._input_to_name = {}
        self._output_to_name = {}
        self._default_drivers = {}
        self._set_definition_attr("primitive", True)
        self._set_definition_attr(_ISWHEN_KEY, True)
        self._set_definition_attr("_builder_", self)
        self._is_when_builder_ = True

    @property
    def default_drivers(self) -> Dict[Type, Type]:
        return self._default_drivers.copy()

    @property
    def input_to_index(self) -> Dict[Type, int]:
        return self._input_to_index.copy()

    @property
    def output_to_index(self) -> Dict[Type, int]:
        return self._output_to_index.copy()

    def _generic_add(
            self,
            value,
            value_to_index,
            index_counter,
            name_counter,
            value_to_name,
            name_prefix,
            type_qualifier,
    ):
        if value is None or value in value_to_index:
            return
        value_to_index[value] = next(index_counter)
        port_name = f"{name_prefix}{next(name_counter)}"
        self._add_port(port_name, type_qualifier(type(value).undirected_t))
        with no_when():
            port = getattr(self, port_name)
            if value.is_input() and isinstance(value, Enable):
                value.wire(
                    port,
                    driven_implicitly_by_when=value.driven_implicitly_by_when
                )
            else:
                wire(port, value)
        value_to_name[value] = port_name

    def add_drivee(self, value: Type):
        self._generic_add(
            value,
            self._output_to_index,
            self._output_counter,
            self._drivee_counter,
            self._output_to_name,
            "O",
            Out,
        )

    def add_driver(self, value: Type):
        self._generic_add(
            value,
            self._input_to_index,
            self._input_counter,
            self._driver_counter,
            self._input_to_name,
            "I",
            In,
        )

    def add_condition(self, condition: Optional[Digital]):
        self._generic_add(
            condition,
            self._input_to_index,
            self._input_counter,
            self._condition_counter,
            self._input_to_name,
            "S",
            In,
        )

    def add_default_driver(self, drivee: Type, driver: Type):
        self.add_driver(driver)
        self.add_drivee(drivee)
        self._default_drivers[drivee] = driver

    def remove_default_driver(self, drivee: Type):
        del self._default_drivers[drivee]

    def _update_resolved_refs(self):
        """When a value driven by a when is resolved, its children are added as
        outputs.  The user may have referenced the unresolved when output using
        .value(), so at this point, we update all the references by checking if
        any resolved outputs are being used to drive values.
        """
        for value, name in self._output_to_name.items():
            if not isinstance(value.name, DerivedRef):
                continue
            if value.name._parent not in self._output_to_name:
                continue
            # Found resolved port, rewire to child output.
            port = getattr(self, self._output_to_name[value.name._parent])
            for drivee in port[value.name.index].driving():
                drivee.rewire(getattr(self, name))

    def _finalize(self):
        self._update_resolved_refs()
        # Detect latches which would be inferred from the context of the when
        # block.
        latches = find_inferred_latches(self.block)
        # NOTE(rsetaluri): These passes should ideally be done after circuit
        # creation. However, it is quite unwieldy, so we opt to do it in this
        # builder's finalization, although it is a bit "hacky". Ideally, this
        # builder should be completely agnostic of anything it is connected to,
        # and a global (post-processing) pass should be used instead.
        # NOTE(leonardt): We only add default drivers if a value is inferred to
        # be a latch, otherwise we skip this implicit logic so the output is
        # cleaner (e.g. if a register is assigned in every case, we don't need
        # the default value).
        _add_default_drivers_to_memory_ports(self, latches)
        _add_default_drivers_to_register_inputs(self, latches)

        if latches:
            raise InferredLatchError(latches)

    @property
    def block(self) -> WhenBlock:
        return self._block


def is_when_builder(builder):
    return isinstance(builder, WhenBuilder)



from typing import List, Tuple

from magma.backend.mlir.magma_common import (
    value_or_value_wrapper_to_tree as magma_value_or_value_wrapper_to_tree
)
from magma.ref import TupleRef, ArrayRef
from magma.view import PortView


def _get_path(tree, node):
    predecessors = tree.predecessors(node)
    try:
        parent = next(predecessors)
    except StopIteration:
        return [node]
    try:
        next(predecessors)
        raise RuntimeError()
    except StopIteration:
        pass
    return _get_path(tree, parent) + [node]


def _path_to_string(path, seperator):
    path = [str(path[0])] + [str(v.name.index) for v in path[1:]]
    return seperator.join(path)


def _get_path_string(tree, node, seperator):
    path = _get_path(tree, node)
    return _path_to_string(path, seperator)


def _get_leaf_descendants(tree, node, include_self=False):
    succ = None
    for succ in tree.successors(node):
        yield from _get_leaf_descendants(tree, succ, include_self=True)
    is_leaf = succ is None
    if is_leaf and include_self:
        yield node


def _find(lst, elt):
    for lst_elt in lst:
        if lst_elt is elt:
            return elt
    return None


def _ascend_to_leaf(value, leaves):
    leaf = _find(leaves, value)
    if leaf is not None:
        return [leaf]
    ref = value.name
    if isinstance(ref, TupleRef):
        path = _ascend_to_leaf(ref.tuple, leaves)
        return path + [f".{ref.index}"]
    if isinstance(ref, ArrayRef):
        path = _ascend_to_leaf(ref.array, leaves)
        return path + [f"[{ref.index}]"]
    raise TypeError(ref, value)


def get_xmr_paths(ctx: 'HardwareModule', xmr: PortView) -> List[Tuple[str]]:
    # We obtain the path to @xmr as follows. (Note that this function only
    # computes the path for the value itself, i.e. the path to the instance in
    # the hierarchy should be computed by the caller.)
    #
    # First, we consider the nested type of the value of which the referred port
    # is a part, as a tree, where array and tuple types result in branches. For
    # example, if the value referred to is `x.y[0].z`, then we build the tree
    # for `x`. The tree construction depends on whether or not we flatten
    # tuples. For example, if we don't flatten tuples, then Tuples are leaves;
    # otherwise, their children are expanded as branches.
    #
    # Next we consider 2 cases:
    # (1) The value referred to is explicitly a node in the constructed tree.
    #     From here, we consider 2 subcases:
    #     (1a) The value referred to is a leaf. In this case, the path is simply
    #          the path to this leaf joined by "_" (since the branches represent
    #          flattening of tuples and arrays).
    #     (1b) Otherwise, we need to consider all leaves which are descendants
    #          of the value referred to. For each such leaf, we compute the path
    #          as in (1a), and return a list of these paths.
    # (2) The value referred to is *not* a node in the constructed tree. This
    #     may happen if the value referred to is a child of a tuple or array
    #     that is not flattened. In this case we need to find the leaf for which
    #     the value referred to is an ancestor. There should be exactly one such
    #     leaf. Then we compute the path as the path to the leaf as computed in
    #     (1a) (i.e. joined with "_"). Then we concatenate the path from this
    #     leaf to the value referred to, joined with ".", since these are not
    #     flattened.
    root = xmr.port.name.root().value()
    tree = magma_value_or_value_wrapper_to_tree(
        root, flatten_all_tuples=ctx.opts.flatten_all_tuples)
    assert tree.has_node(root)

    # (1)
    if tree.has_node(xmr.port):  # visited
        path = _get_path_string(tree, xmr.port, "_")
        # (1a)
        if tree.out_degree(xmr.port) == 0:  # is leaf
            return [(path,)]
        # (1b)
        leaves = _get_leaf_descendants(tree, xmr.port)
        leaves = sorted(leaves, key=lambda n: tree.nodes[n]["index"])
        return [
            (_get_path_string(tree, leaf, "_"),)
            for leaf in leaves
        ]
    # (2)
    leaves = list(_get_leaf_descendants(tree, root, include_self=True))
    leaf, *path = _ascend_to_leaf(xmr.port, leaves)
    path = _get_path_string(tree, leaf, "_") + "".join(path)
    return [(path,)]



import dataclasses
from typing import List, Optional

from magma.backend.mlir.hw import hw
from magma.backend.mlir.mlir import (
    MlirDialect, MlirOp, MlirBlock, MlirValue, MlirSymbol,
    begin_dialect, end_dialect)
from magma.backend.mlir.mlir_printer_utils import print_names, print_types
from magma.backend.mlir.print_opts import PrintOpts
from magma.backend.mlir.printer_base import PrinterBase


sv = MlirDialect("sv")
begin_dialect(sv)


@dataclasses.dataclass
class RegOp(MlirOp):
    results: List[MlirValue]
    name: Optional[str] = None

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(f" = sv.reg ")
        if self.name is not None:
            printer.print(f"{{name = \"{self.name}\"}} ")
        printer.print(": ")
        print_types(self.results, printer)


@dataclasses.dataclass
class ReadInOutOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(f" = sv.read_inout ")
        print_names(self.operands, printer)
        printer.print(" : ")
        print_types(self.operands, printer)


@dataclasses.dataclass
class AssignBaseOp(MlirOp):
    operands: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        printer.print(f"sv.{self.op_name} ")
        print_names(self.operands, printer)
        printer.print(" : ")
        print_types(self.operands[1], printer)


@dataclasses.dataclass
class AssignOp(AssignBaseOp):
    operands: List[MlirValue]
    op_name = "assign"


@dataclasses.dataclass
class PAssignOp(AssignBaseOp):
    operands: List[MlirValue]
    op_name = "passign"


@dataclasses.dataclass
class BPAssignOp(AssignBaseOp):
    operands: List[MlirValue]
    op_name = "bpassign"


@dataclasses.dataclass
class AlwaysFFOp(MlirOp):
    operands: List[MlirValue]
    clock_edge: str
    reset_type: str = None
    reset_edge: str = None

    def __post_init__(self):
        self._body_block = self.new_region().new_block()
        if self.reset_type is None:
            return
        self._reset_block = self.new_region().new_block()

    @property
    def body_block(self) -> MlirBlock:
        return self._body_block

    @property
    def reset_block(self) -> MlirBlock:
        return self._reset_block

    def print(self, printer: PrinterBase, opts: PrintOpts):
        printer.print(f"sv.alwaysff({self.clock_edge} ")
        print_names(self.operands[0], printer)
        printer.print(") {")
        printer.flush()
        printer.push()
        self.body_block.print(printer, opts)
        printer.pop()
        printer.print("}")
        if self.reset_type is None:
            printer.flush()
            return
        printer.print(f" ({self.reset_type} : {self.reset_edge} ")
        print_names(self.operands[1], printer)
        printer.print(") {")
        printer.flush()
        printer.push()
        self.reset_block.print(printer, opts)
        printer.pop()
        printer.print_line("}")

    def print_op(self, printer: PrinterBase):
        raise NotImplementedError()


@dataclasses.dataclass
class AlwaysCombOp(MlirOp):
    def __post_init__(self):
        self._body_block = self.new_region().new_block()

    @property
    def body_block(self) -> MlirBlock:
        return self._body_block

    def print(self, printer: PrinterBase, opts: PrintOpts):
        printer.print("sv.alwayscomb {")
        printer.flush()
        printer.push()
        self.body_block.print(printer, opts)
        printer.pop()
        printer.print_line("}")

    def print_op(self, printer: PrinterBase):
        raise NotImplementedError()


@dataclasses.dataclass
class InitialOp(MlirOp):
    def __post_init__(self):
        self._block = self.new_region().new_block()

    def add_operation(self, operation: MlirOp):
        self._block.add_operation(operation)

    def print_op(self, printer: PrinterBase):
        printer.print("sv.initial")


@dataclasses.dataclass
class WireOp(MlirOp):
    results: List[MlirValue]
    name: Optional[str] = None
    sym: Optional[MlirSymbol] = None

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(" = sv.wire ")
        if self.sym is not None:
            printer.print(f"sym {self.sym.name} ")
        if self.name is not None:
            printer.print(f"{{name=\"{self.name}\"}} ")
        printer.print(": ")
        print_types(self.results, printer)


def _esacpe_string(string: str):
    # NOTE(rsetaluri): This is a hack to "double-escape" escape characters like
    # `\n`, `\t`.
    return (
        repr(string)[1:-1]
        .replace("\"", "\\\"")
        .replace("\\\'", "'")
    )


@dataclasses.dataclass
class VerbatimOp(MlirOp):
    operands: List[MlirOp]
    string: str

    def print_op(self, printer: PrinterBase):
        string = _esacpe_string(self.string)
        printer.print(f"sv.verbatim \"{string}\"")
        if self.operands:
            printer.print(" (")
            print_names(self.operands, printer)
            printer.print(") : ")
            print_types(self.operands, printer)


@dataclasses.dataclass
class VerbatimExprOp(MlirOp):
    operands: List[MlirOp]
    results: List[MlirOp]
    expr: str

    def print_op(self, printer: PrinterBase):
        expr = _esacpe_string(self.expr)
        print_names(self.results, printer)
        printer.print(f" = sv.verbatim.expr \"{expr}\"")
        printer.print(" (")
        print_names(self.operands, printer)
        printer.print(") : (")
        print_types(self.operands, printer)
        printer.print(") -> (")
        print_types(self.results, printer)
        printer.print(")")


@dataclasses.dataclass
class BindOp(MlirOp):
    instance: hw.InnerRefAttr

    def print_op(self, printer: PrinterBase):
        printer.print(f"sv.bind {self.instance.emit()}")


@dataclasses.dataclass
class IfDefOp(MlirOp):
    cond: str

    def __post_init__(self):
        self._then_block = self.new_region().new_block()
        self._else_block = None

    @property
    def then_block(self) -> MlirBlock:
        return self._then_block

    @property
    def else_block(self) -> MlirBlock:
        if self._else_block is None:
            self._else_block = self.new_region().new_block()
        return self._else_block

    def print(self, printer: PrinterBase, opts: PrintOpts):
        printer.print(f"sv.ifdef \"{self.cond}\" {{")
        printer.flush()
        printer.push()
        self._then_block.print(printer, opts)
        printer.pop()
        printer.print("}")
        if self._else_block is None:
            printer.flush()
            return
        printer.print(" else {")
        printer.flush()
        printer.push()
        self._else_block.print(printer, opts)
        printer.pop()
        printer.print_line("}")

    def print_op(self, printer: PrinterBase):
        raise NotImplementedError()


@dataclasses.dataclass
class IfOp(MlirOp):
    operands: List[MlirOp]

    def __post_init__(self):
        self._then_block = self.new_region().new_block()
        self._else_block = None

    @property
    def then_block(self) -> MlirBlock:
        return self._then_block

    @property
    def else_block(self) -> MlirBlock:
        if self._else_block is None:
            self._else_block = self.new_region().new_block()
        return self._else_block

    def print(self, printer: PrinterBase, opts: PrintOpts):
        printer.print(f"sv.if ")
        print_names(self.operands[0], printer)
        printer.print(" {")
        printer.flush()
        printer.push()
        self._then_block.print(printer, opts)
        printer.pop()
        printer.print("}")
        if self._else_block is None:
            printer.flush()
            return
        printer.print(" else {")
        printer.flush()
        printer.push()
        self._else_block.print(printer, opts)
        printer.pop()
        printer.print_line("}")

    def print_op(self, printer: PrinterBase):
        raise NotImplementedError()


@dataclasses.dataclass
class XMROp(MlirOp):
    results: List[MlirOp]
    is_rooted: bool
    path: List[str]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(f" = sv.xmr ")
        if self.is_rooted:
            printer.print(f"isRooted ")
        path = [f"\"{p}\"" for p in self.path]
        path = f"{', '.join(path)}"
        printer.print(f"{path} : ")
        print_types(self.results, printer)


@dataclasses.dataclass
class ArrayIndexInOutOp(MlirOp):
    operands: List[MlirOp]
    results: List[MlirOp]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(f" = sv.array_index_inout ")
        print_names(self.operands[0], printer)
        printer.print("[")
        print_names(self.operands[1], printer)
        printer.print("] : ")
        print_types(self.operands, printer)


end_dialect()



from magma.array import Array
from magma.tuple import Tuple


def make_unconnected_error_str(port):
    format_args = [port]
    error_str = "{}"
    if port.trace() is not None:
        error_str += ": Connected"
    elif isinstance(port, (Tuple, Array)):
        child_str = ""
        for child in port:
            child, child_args = make_unconnected_error_str(child)
            format_args.extend(child_args)
            child = "\n    ".join(child.splitlines())
            child_str += f"\n    {child}"
        if "Connected" not in child_str:
            # Handle case when no children are connected (simplify)
            error_str += ": Unconnected"
        else:
            error_str += child_str
    elif port.trace() is None:
        error_str += ": Unconnected"
    return error_str, format_args


class MagmaCompileException(Exception):
    pass


class UnconnectedPortException(MagmaCompileException):
    def __init__(self, port):
        msg = f"Found unconnected port: {port.debug_name}\n"
        error_str, format_args = make_unconnected_error_str(port)
        msg += error_str.format(*(arg.debug_name for arg in format_args))
        super().__init__(msg)



import os
import magma.ast_utils as ast_utils
import ast
import astor
from .verilog_utils import CollectInitialWidthsAndTypes, PromoteWidths, \
    RemoveBits
import magma as m
import kratos
from magma.config import get_debug_mode


class SpecializeConstantInts(ast.NodeTransformer):
    def __init__(self, defn_env):
        self.defn_env = defn_env

    def visit_Name(self, node):
        if node.id in self.defn_env:
            value = self.defn_env[node.id]
            if isinstance(value, int):
                return ast.Num(value)
        return node


class RewriteSelfAttributes(ast.NodeTransformer):
    def __init__(self, initial_value_map):
        self.initial_value_map = initial_value_map
        self.calls_seen = []

    def visit_Attribute(self, node):
        if isinstance(node.value, ast.Name) and node.value.id == "self":
            return ast.Name(f"self_{node.attr}_I", node.ctx)
        return self.generic_visit(node)

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute) and \
                isinstance(node.func.value, ast.Name) and \
                node.func.value.id == "self":
            attr = node.func.attr
            assert attr in self.initial_value_map, \
                "Reference to self that was not initialized"
            ret = []
            ports = self.initial_value_map[attr][3].interface.inputs()
            for name, value in zip(ports, node.args):
                name = magma_value(name)
                ret.append(ast.parse(
                    f"self_{attr}_{name} = {astor.to_source(value).rstrip()}"
                ).body[0])
            self.calls_seen.extend(ret)
            func = astor.to_source(node.func).rstrip()
            outputs = self.initial_value_map[attr][3].interface.outputs()
            if len(outputs) == 1:
                name = outputs[0]
                name = magma_value(name)
                return ast.Name(f"self_{attr}_{name}", ast.Load())
            else:
                assert outputs, "Expected module with at least one output"
                return ast.Tuple([ast.Name(f"self_{attr}_{magma_value(output)}",
                                           ast.Load()) for output in outputs],
                                 ast.Load())
        elif (isinstance(node.func, ast.Attribute) and
                isinstance(node.func.value, ast.Attribute) and
                isinstance(node.func.value.value, ast.Name) and
                node.func.value.value.id == "self" and
                node.func.attr == "prev"):
            return ast.Name(f"self_{node.func.value.attr}_O", ast.Load())
        else:
            self.generic_visit(node)
        return node

    def visit_If(self, node):
        node.test = self.visit(node.test)
        for attr in ["body", "orelse"]:
            new_body = []
            for item in getattr(node, attr):
                new_body.append(self.visit(item))
            if self.calls_seen:
                new_body = self.calls_seen + new_body
                self.calls_seen = []
            setattr(node, attr, new_body)
        return node


def get_initial_value_map(init_func, defn_env):
    """
    Parses __init__ funciton of the form
        def __init__(self):
            self.x = m.bits(0, 2)
            self.y = m.bits(0, 4)
    Returns a map from instance attributes to their initial values
    For the above example,
        {'x': <_ast.Call object at 0x109c17fd0>,
         'y': <_ast.Call object at 0x109c17470>}
    TODO: Should we allow arbitrary code in the definition? For now, the user
    can write any meta code outside the class definition, which makes the
    design of this simpler
    """
    initial_value_map = {}
    if init_func is object.__init__:
        # Handle case when no __init__
        return initial_value_map
    init_def = ast_utils.get_ast(init_func).body[0]
    # init_def = ExecuteEscapedPythonExpressions(defn_env).visit(init_def)
    init_def = SpecializeConstantInts(defn_env).visit(init_def)
    for stmt in init_def.body:
        # We only support basic assignments of the form
        #     self.x: m.Bits(2) = m.bits(0, 2)
        assert isinstance(stmt, ast.AnnAssign)
        assert isinstance(stmt.target, ast.Attribute)
        assert isinstance(stmt.target.value, ast.Name)
        assert stmt.target.value.id == "self"

        # TODO: Should we deal with multiple assignments? For now we take the
        # last one
        class Tmp(m.Circuit):
            eval_type = eval(astor.to_source(stmt.annotation).rstrip(),
                             defn_env)
            eval_value = eval(astor.to_source(stmt.value).rstrip(), defn_env)
        initial_value_map[stmt.target.attr] = (stmt.value, stmt.annotation,
                                               Tmp.eval_type,
                                               Tmp.eval_value)
    return initial_value_map


class ProcessNames(ast.NodeTransformer):
    def __init__(self, inputs, outputs, names):
        self.names = list(inputs.keys()) + list(outputs.keys()) + list(names)

    def visit_Name(self, node):
        if node.id in self.names:
            return ast.parse(f"self.{node.id}").body[0].value
        return node


class ProcessReturns(ast.NodeTransformer):
    def __init__(self, outputs):
        self.len_outputs = len(outputs)

    def visit_Return(self, node):
        if self.len_outputs > 1:
            assert isinstance(node.value, ast.Tuple)
            return [
                ast.Assign(
                    [ast.parse(f"self.O{i}").body[0].value],
                    node.value.elts[0],
                    lineno=node.lineno
                )
                for i in range(self.len_outputs)
            ]
        else:
            return ast.Assign(
                [ast.parse(f"self.O").body[0].value],
                node.value,
                lineno=node.lineno
            )


def to_type_str(type_):
    if issubclass(type_, m.Digital):
        return "bit"
    elif issubclass(type_, m.UInt):
        return "uint"
    elif issubclass(type_, m.Bits):
        return "bits"
    else:
        raise NotImplementedError(type_)


class NameCollector(ast.NodeVisitor):
    def __init__(self, ctx):
        self.names = set()
        self.ctx = ctx

    def visit_Name(self, node):
        if self.ctx is None or isinstance(node.ctx, self.ctx):
            self.names.add(node.id)


def collect_names(tree, ctx=None):
    visitor = NameCollector(ctx)
    visitor.visit(tree)
    return visitor.names


def get_io_width(type_):
    if type_ is m.Bit:
        return None
    elif isinstance(type_, m.ArrayMeta):
        if isinstance(type_.T, m.ArrayMeta):
            elem_width = get_io_width(type_.T)
            if isinstance(elem_width, tuple):
                return (type_.N, ) + elem_width
            else:
                return (type_.N, elem_width)
        else:
            return type_.N
    else:
        raise NotImplementedError(type_)


def get_length(t):
    try:
        # assume flattened type, so it must be an array of Bits or a Bit
        return len(t)
    except Exception:
        # Bit is treated as length 1
        return 1


def build_kratos_debug_info(circuit, is_top):
    inst_to_defn_map = {}
    for instance in circuit.instances:
        instance_inst_to_defn_map = \
            build_kratos_debug_info(type(instance), is_top=False)
        for k, v in instance_inst_to_defn_map.values():
            key = instance.name + "." + k
            if is_top:
                key = circuit.name + "." + key
            inst_to_defn_map[key] = v
        inst_name = instance.name
        if is_top:
            inst_name = circuit.name + "." + instance.name
        if instance.kratos is not None:
            inst_to_defn_map[inst_name] = instance.kratos
    return inst_to_defn_map


def process_func(defn_env, fn, circ_name, registers=None, debug=False):
    tree = ast_utils.get_ast(fn).body[0]
    # TODO: Flatten types pass
    inputs = {}
    outputs = {}
    width_table = {}
    type_table = {}
    # determine if it's sequential or combinational
    combinational = registers is None
    if not combinational:
        tree = RewriteSelfAttributes(registers).visit(tree)
        for name, info in registers.items():
            width = 1 if isinstance(info[3], m.DigitalMeta) else len(info[3])
            width_table["self_" + name + "_I"] = width
            type_table["self_" + name + "_I"] = to_type_str(info[2])
            width_table["self_" + name + "_O"] = width
            type_table["self_" + name + "_O"] = to_type_str(info[2])
        for name in reversed(tuple(registers.keys())):
            tree.body.insert(
                0, ast.parse(f"self_{name}_I = self_{name}_O")
            )
    else:
        registers = {}
    for arg in tree.args.args:
        if arg.arg == "self":
            continue
        type_ = eval(astor.to_source(arg.annotation), defn_env)
        inputs[arg.arg] = get_length(type_)
        width_table[arg.arg] = get_io_width(type_)
        type_table[arg.arg] = to_type_str(type_)
    if isinstance(tree.returns, ast.Tuple):
        for i, elt in enumerate(tree.returns.elts):
            type_ = eval(astor.to_source(elt), defn_env)
            outputs[f"O{i}"] = get_length(type_)
            width_table[f"O{i}"] = get_io_width(type_)
            type_table[f"O{i}"] = to_type_str(type_)
    else:
        type_ = eval(astor.to_source(tree.returns), defn_env)
        outputs["O"] = get_length(type_)
        width_table["O"] = get_io_width(type_)
        type_table["O"] = to_type_str(type_)

    names = [arg.arg for arg in tree.args.args if arg.arg != "self"]
    types = [arg.annotation for arg in tree.args.args if arg.arg != "self"]
    IO = []
    for name, type_ in zip(names, types):
        IO += [name, m.In(eval(compile(ast.Expression(type_), "",
                                       mode="eval")))]
    if isinstance(tree.returns, ast.Tuple):
        for i, elt in enumerate(tree.returns.elts):
            IO += [f"O{i}", m.Out(eval(compile(ast.Expression(elt), "",
                                               mode="eval")))]
    else:
        IO += [f"O", m.Out(eval(compile(ast.Expression(tree.returns), "",
                                        mode="eval")))]
    if registers:
        IO += m.ClockInterface(has_async_reset=True)

    names = collect_names(tree, ctx=ast.Store) | \
        set(f"self_{name}_I" for name in registers) | \
        set(f"self_{name}_O" for name in registers)

    for io_port in inputs.keys() | outputs.keys():
        if io_port in names:
            names.remove(io_port)

    CollectInitialWidthsAndTypes(width_table, type_table, defn_env).visit(tree)
    PromoteWidths(width_table, type_table).visit(tree)
    tree = ProcessNames(inputs, outputs, names).visit(tree)
    tree = ProcessReturns(outputs).visit(tree)
    tree = RemoveBits().visit(tree)
    tree.args.args = [ast.arg("self", None)]
    tree.decorator_list = []
    tree.returns = None

    # obtain the filename and line number information
    fn_ln = kratos.pyast.get_fn(fn), kratos.pyast.get_ln(fn)

    print(astor.to_source(tree))

    class Module(kratos.Generator):
        def __init__(self):
            super().__init__(circ_name, get_debug_mode() or debug)
            for key, value in inputs.items():
                setattr(self, key, self.input(key, value))
            for key, value in outputs.items():
                setattr(self, key, self.output(key, value))
            for name in names:
                setattr(self, name, self.var(name, width_table[name]))

            block = self.add_code(tree, fn_ln=fn_ln)
            if not combinational:
                # TODO: add option to do reset low
                reset_high = True
                self.CLK = self.clock("CLK")
                self.ASYNCRESET = self.reset("ASYNCRESET")
                sensitivity = [(kratos.posedge, self.CLK),
                               (kratos.posedge
                                if reset_high else kratos.negedge,
                                self.ASYNCRESET)]
                seq = self.sequential(*sensitivity)
                # add an if statement
                if reset_high:
                    if_ = seq.if_(self.ASYNCRESET)
                else:
                    if_ = seq.if_(self.ASYNCRESET.r_not())
                for reg, values in registers.items():
                    try:
                        init = int(values[3])
                    except Exception:
                        raise NotImplementedError(values[3])
                    if_.then_(self.vars[f"self_{reg}_O"].assign(init))
                for n in registers:
                    if_.else_(self.vars[f"self_{n}_O"].assign(
                        self.vars[f"self_{n}_I"]))

            # set locals, which is the inputs and outputs
            for key in inputs:
                block.stmt().add_scope_variable(key, key, True)
            for key in names:
                if "self_" in key:
                    continue
                block.stmt().add_scope_variable(key, key, True)
            if registers:
                for key in registers:
                    block.stmt().add_scope_variable(key, f"self_{key}_I", True)

    os.makedirs(".magma", exist_ok=True)
    filename = f".magma/{circ_name}-kratos.sv"
    mod = Module()

    kratos.verilog(mod, filename=filename,
                   insert_debug_info=get_debug_mode() or debug)
    io_dict = {key: value for key, value in zip(IO[::2], IO[1::2])}

    class _Defn(m.Circuit):
        name = circ_name
        io = m.IO(**io_dict)
        kratos = mod
    with open(filename, 'r') as f:
        _Defn.verilogFile = f.read()
    return _Defn


def combinational_to_verilog(debug=False):
    @ast_utils.inspect_enclosing_env
    def wrapped(defn_env, fn):
        return lambda *args: process_func(defn_env, fn, fn.__name__,
                                          debug=debug)()(*args)
    return wrapped


def sequential_to_verilog(async_reset, debug=False):
    if not async_reset:
        raise NotImplementedError()

    @ast_utils.inspect_enclosing_env
    def wrapped(defn_env, cls):
        initial_value_map = get_initial_value_map(cls.__init__, defn_env)
        return process_func(defn_env, cls.__call__, cls.__name__,
                            initial_value_map, debug=debug)
    return wrapped



import abc
import contextlib
import dataclasses
import functools
import itertools
import pathlib
from typing import Any, List, Mapping, Optional, Tuple, Union
import weakref

from magma.array import Array, ArrayMeta
from magma.backend.mlir.build_magma_graph import (
    BuildMagmaGrahOpts, build_magma_graph
)
from magma.backend.mlir.builtin import builtin
from magma.backend.mlir.comb import comb
from magma.backend.mlir.common import wrap_with_not_implemented_error
from magma.backend.mlir.compile_to_mlir_opts import CompileToMlirOpts
from magma.backend.mlir.errors import MlirCompilerInternalError
from magma.backend.mlir.graph_lib import Graph
from magma.backend.mlir.hw import hw
from magma.backend.mlir.magma_common import (
    ModuleLike as MagmaModuleLike,
    ValueWrapper as MagmaValueWrapper,
    InstanceWrapper as MagmaInstanceWrapper,
    value_or_type_to_string as magma_value_or_type_to_string,
    visit_value_or_value_wrapper_by_direction as
    visit_magma_value_or_value_wrapper_by_direction,
)
from magma.backend.mlir.mlir import (
    MlirType, MlirValue, MlirSymbol, MlirAttribute, MlirBlock, push_block
)
from magma.backend.mlir.scoped_name_generator import ScopedNameGenerator
from magma.backend.mlir.sv import sv
from magma.backend.mlir.xmr_utils import get_xmr_paths
from magma.bind2 import is_bound_instance
from magma.bit import Bit
from magma.bits import Bits, BitsMeta
from magma.bitutils import clog2
from magma.circuit import AnonymousCircuitType, CircuitKind, DefineCircuitKind
from magma.clock import Reset, ResetN, AsyncReset, AsyncResetN
from magma.common import filter_by_key, sort_by_value
from magma.compile_guard import get_compile_guard_data
from magma.digital import Digital, DigitalMeta
from magma.inline_verilog_expression import InlineVerilogExpression
from magma.is_definition import isdefinition
from magma.is_primitive import isprimitive
from magma.linking import (
    has_any_linked_modules,
    get_linked_modules,
    has_default_linked_module,
    get_default_linked_module,
)
from magma.primitives.mux import Mux
from magma.primitives.register import Register
from magma.primitives.when import iswhen
from magma.primitives.wire import Wire
from magma.primitives.xmr import XMRSink, XMRSource
from magma.protocol_type import magma_value as get_magma_value
from magma.t import Kind, Type
from magma.tuple import TupleMeta, Tuple as m_Tuple
from magma.value_utils import make_selector, TupleSelector, ArraySelector
from magma.view import PortView


MlirValueList = List[MlirValue]


def _assert_false(*args, **kwargs):
    assert False


def _get_defn_or_decl_output_name(
        ctx: 'HardwareModule', defn_or_decl: CircuitKind) -> str:
    metadata = defn_or_decl.coreir_metadata
    try:
        return metadata["verilog_name"]
    except KeyError:
        pass
    name = defn_or_decl.name
    if ctx.opts.user_namespace is not None:
        name = ctx.opts.user_namespace + f"_{name}"
    if ctx.opts.verilog_prefix is not None:
        name = ctx.opts.verilog_prefix + f"{name}"
    return name


def _make_compile_guard_block(compile_guard: Mapping) -> MlirBlock:
    if_def = sv.IfDefOp(compile_guard["condition_str"])
    if compile_guard["type"] == "defined":
        return if_def.then_block
    assert compile_guard["type"] == "undefined"
    return if_def.else_block


def _maybe_set_location_info(op, debug_info):
    if debug_info is None:
        return
    loc = builtin.FileLineColLoc(debug_info.filename, debug_info.lineno, 0)
    op.location = loc


@wrap_with_not_implemented_error
def parse_reset_type(T: Kind) -> Tuple[str, str]:
    if T is Reset:
        return "syncreset", "posedge"
    if T is ResetN:
        return "syncreset", "negedge"
    if T is AsyncReset:
        return "asyncreset", "posedge"
    if T is AsyncResetN:
        return "asyncreset", "negedge"


@wrap_with_not_implemented_error
@functools.lru_cache()
def magma_type_to_mlir_type(type: Kind) -> MlirType:
    type = type.undirected_t
    if issubclass(type, Digital):
        return builtin.IntegerType(1)
    if issubclass(type, Bits):
        return builtin.IntegerType(type.N)
    if issubclass(type, Array):
        if issubclass(type.T, Bit):
            return magma_type_to_mlir_type(Bits[type.N])
        return hw.ArrayType((type.N,), magma_type_to_mlir_type(type.T))
    if issubclass(type, m_Tuple):
        fields = {str(k): magma_type_to_mlir_type(t)
                  for k, t in type.field_dict.items()}
        return hw.StructType(tuple(fields.items()))


@wrap_with_not_implemented_error
@functools.lru_cache()
def python_type_to_mlir_type(type_: type) -> MlirType:
    # NOTE(rsetaluri): We only support integer attribtue types right now. All
    # integer parameter types are assumed to be int32's.
    if type_ is int:
        return builtin.IntegerType(32)


@wrap_with_not_implemented_error
def python_value_to_mlir_attribtue(value: Any) -> MlirAttribute:
    # NOTE(rsetaluri): We only support integer attribute types right now.
    if isinstance(value, int):
        return builtin.IntegerAttr(value)


def get_module_interface(
        module: MagmaModuleLike, ctx) -> Tuple[MlirValueList, MlirValueList]:
    operands = []
    results = []
    for port in module.interface.ports.values():
        port = get_magma_value(port)
        visit_magma_value_or_value_wrapper_by_direction(
            port,
            lambda p: operands.append(ctx.get_or_make_mapped_value(p)),
            lambda p: results.append(ctx.get_or_make_mapped_value(p)),
            flatten_all_tuples=ctx.opts.flatten_all_tuples,
        )
    return operands, results


def make_mux(
        ctx: 'HardwareModule',
        data: List[MlirValue],
        select: MlirValue,
        result: MlirValue):
    if ctx.opts.extend_non_power_of_two_muxes:
        closest_power_of_two_size = 2 ** clog2(len(data))
        if closest_power_of_two_size != len(data):
            extension = [data[0]] * (closest_power_of_two_size - len(data))
            data = extension + data
    mlir_type = hw.ArrayType((len(data),), data[0].type)
    array = ctx.new_value(mlir_type)
    hw.ArrayCreateOp(
        operands=data,
        results=[array])
    hw.ArrayGetOp(
        operands=[array, select],
        results=[result])


def get_register_data_and_init(
        ctx: 'HardwareModule', I: Type, operands: List[MlirValue], init: Type):
    T_out, data_out, init_out = [], [], []

    def _visit_input(i):
        T_out.append(type(i))
        data_out.append(operands.pop(0))
        init_out.append(make_selector(i).select(init))

    visit_magma_value_or_value_wrapper_by_direction(
        I,
        _visit_input,
        _visit_input,
        flatten_all_tuples=ctx.opts.flatten_all_tuples
    )
    return T_out, data_out, init_out


def make_hw_param_decl(name: str, value: Any):
    type_ = python_type_to_mlir_type(type(value))
    value = python_value_to_mlir_attribtue(value)
    return hw.ParamDeclAttr(name, type_, value)


def make_hw_instance_op(
        operands: MlirValueList,
        results: MlirValueList,
        name: str,
        module: hw.ModuleOp,
        parameters: Mapping[str, Any] = dict(),
        sym: Optional[MlirSymbol] = None,
        compile_guard: Optional[Mapping] = None,
        attrs: Optional[Mapping] = None,
) -> hw.InstanceOp:
    if compile_guard is not None:
        ctx = push_block(_make_compile_guard_block(compile_guard))
    else:
        ctx = contextlib.nullcontext()
    parameters = [
        make_hw_param_decl(name, value) for name, value in parameters.items()
    ]
    with ctx:
        op = hw.InstanceOp(
            name=name,
            module=module,
            operands=operands,
            results=results,
            parameters=parameters,
            sym=sym)
    if attrs is not None:
        op.attr_dict.update(attrs)
    return op


def resolve_xmr(ctx: 'HardwareModule', xmr: PortView):
    assert isinstance(xmr, PortView)
    mlir_type = magma_type_to_mlir_type(type(xmr)._to_magma_())
    in_out = ctx.new_value(hw.InOutType(mlir_type))
    sv.XMROp(is_rooted=False, path=list(xmr.path()), results=[in_out])
    value = ctx.new_value(mlir_type)
    sv.ReadInOutOp(operands=[in_out], results=[value])
    return value


@dataclasses.dataclass(frozen=True)
class ModuleWrapper:
    module: MagmaModuleLike
    operands: MlirValueList
    results: MlirValueList

    @staticmethod
    def make(module: MagmaModuleLike, ctx) -> 'ModuleWrapper':
        operands, results = get_module_interface(module, ctx)
        return ModuleWrapper(module, operands, results)


class ModuleVisitor:

    class VisitError(RuntimeError):
        pass

    class RevisitedModuleError(VisitError):
        pass

    def __init__(self, graph: Graph, ctx):
        self._graph = graph
        self._ctx = ctx
        self._visited = set()

    @functools.lru_cache()
    def make_constant(
            self, T: Kind, value: Optional[Any] = None) -> MlirValue:
        result = self._ctx.new_value(T)
        if isinstance(T, (DigitalMeta, BitsMeta)):
            value = value if value is not None else 0
            hw.ConstantOp(value=int(value), results=[result])
            return result
        if isinstance(T, ArrayMeta):
            value = value if value is not None else (None for _ in range(T.N))
            operands = [self.make_constant(T.T, v) for v in value]
            hw.ArrayCreateOp(operands=operands, results=[result])
            return result
        if isinstance(T, TupleMeta):
            fields = T.field_dict.items()
            value = value if value is not None else {k: None for k, _ in fields}
            operands = [self.make_constant(t, value[k]) for k, t in fields]
            hw.StructCreateOp(operands=operands, results=[result])
            return result
        raise TypeError(T)

    @wrap_with_not_implemented_error
    def visit_coreir_mem(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert defn.coreir_name == "mem"
        # TODO(rsetaluri): Add support for initialization.
        if defn.coreir_genargs["has_init"]:
            raise NotImplementedError("coreir.mem init not supported")
        width = defn.coreir_genargs["width"]
        depth = defn.coreir_genargs["depth"]
        raddr, waddr, wdata, clk, wen = module.operands
        rdata = module.results[0]
        elt_type = hw.InOutType(builtin.IntegerType(width))
        reg_type = hw.InOutType(hw.ArrayType((depth,), elt_type.T))
        reg = self._ctx.new_value(reg_type)
        sv.RegOp(name=inst.name, results=[reg])
        # Register read logic.
        read = self._ctx.new_value(elt_type)
        sv.ArrayIndexInOutOp(operands=[reg, raddr], results=[read])
        sv.ReadInOutOp(operands=[read], results=[rdata])
        # Register write logic.
        write = self._ctx.new_value(elt_type)
        sv.ArrayIndexInOutOp(operands=[reg, waddr], results=[write])
        always = sv.AlwaysFFOp(operands=[clk], clock_edge="posedge").body_block
        with push_block(always):
            with push_block(sv.IfOp(operands=[wen]).then_block):
                sv.PAssignOp(operands=[write, wdata])
        return True

    @wrap_with_not_implemented_error
    def visit_coreir_not(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert defn.coreir_name == "not"
        neg_one = self.make_constant(type(inst.I), -1)
        comb.BaseOp(
            op_name="xor",
            operands=[neg_one, module.operands[0]],
            results=module.results)
        return True

    @wrap_with_not_implemented_error
    def visit_coreir_reg(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert defn.coreir_name == "reg" or defn.coreir_name == "reg_arst"
        reg = self._ctx.new_value(
            hw.InOutType(magma_type_to_mlir_type(type(defn.O))))
        sv.RegOp(name=inst.name, results=[reg])
        clock_edge = "posedge"
        has_reset = (defn.coreir_name == "reg_arst")
        operands = [module.operands[1]]
        attrs = dict(clock_edge=clock_edge)
        if has_reset:
            reset_type = "asyncreset"
            arst_posedge = defn.coreir_configargs["arst_posedge"]
            reset_edge = "posedge" if arst_posedge else "negedge"
            operands.append(module.operands[2])
            attrs.update(dict(reset_type=reset_type, reset_edge=reset_edge))
        always = sv.AlwaysFFOp(operands=operands, **attrs)
        init = defn.coreir_configargs["init"].value
        const = self.make_constant(type(defn.I), init)
        with push_block(always.body_block):
            sv.PAssignOp(operands=[reg, module.operands[0]])
        if has_reset:
            always.operands.append(module.operands[1])
            with push_block(always.reset_block):
                sv.PAssignOp(operands=[reg, const])
        if not self._ctx.opts.disable_initial_blocks:
            with push_block(sv.InitialOp()):
                sv.BPAssignOp(operands=[reg, const])
        sv.ReadInOutOp(operands=[reg], results=module.results.copy())
        return True

    @wrap_with_not_implemented_error
    def visit_coreir_reduce(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert (defn.coreir_name in ("orr", "andr", "xorr"))
        size = len(defn.I)
        if defn.coreir_name == "orr":
            const = self.make_constant(type(defn.I), value=0)
            comb.ICmpOp(
                predicate="ne",
                operands=[module.operands[0], const],
                results=module.results)
        elif defn.coreir_name == "andr":
            const = self.make_constant(type(defn.I), value=-1)
            comb.ICmpOp(
                predicate="eq",
                operands=[module.operands[0], const],
                results=module.results)
        elif defn.coreir_name == "xorr":
            comb.ParityOp(
                operands=module.operands,
                results=module.results)
        return True

    @wrap_with_not_implemented_error
    def visit_coreir_wire(self, module: ModuleWrapper) -> bool:
        inst = module.module

        def _visit(value, counter):
            index = next(counter)
            mlir_type = hw.InOutType(module.operands[index].type)
            wire = self._ctx.new_value(mlir_type)
            name = inst.name
            # TODO(rsetaluri): Making a selector and inspecting it in order to
            # find the name is a kind of hacky way to get the qualified name of
            # the sub-field of the value. Instead we should have a common API
            # (what "qualifiedname" really should be...) to obtain this name.
            selector = make_selector(value)
            if isinstance(selector, (TupleSelector, ArraySelector)):
                name += str(selector).replace(".", "_")
            else:
                assert index == 0
            sym = self._ctx.parent.get_or_make_mapped_symbol(
                value, name=f"{self._ctx.name}.{name}", force=True
            )
            sv.WireOp(results=[wire], name=name, sym=sym)
            sv.AssignOp(operands=[wire, module.operands[index]])
            sv.ReadInOutOp(operands=[wire], results=[module.results[index]])

        counter = itertools.count()
        visit_magma_value_or_value_wrapper_by_direction(
            inst.I,
            lambda v: _visit(v, counter),
            _assert_false,
            flatten_all_tuples=self._ctx.opts.flatten_all_tuples,
        )
        return True

    @wrap_with_not_implemented_error
    def visit_coreir_primitive(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert (defn.coreir_lib == "coreir" or defn.coreir_lib == "corebit")
        if defn.coreir_name == "mem":
            return self.visit_coreir_mem(module)
        if defn.coreir_name == "not":
            return self.visit_coreir_not(module)
        if defn.coreir_name in (
                "eq", "ult", "eq", "ne", "slt", "sle", "sgt", "sge", "ult",
                "ule", "ugt", "uge",
        ):
            comb.ICmpOp(
                predicate=defn.coreir_name,
                operands=module.operands,
                results=module.results)
            return True
        if defn.coreir_name in ("reg", "reg_arst"):
            return self.visit_coreir_reg(module)
        if defn.coreir_name in ("orr", "andr", "xorr"):
            return self.visit_coreir_reduce(module)
        if defn.coreir_name == "wire":
            return self.visit_coreir_wire(module)
        if defn.coreir_name == "wrap":
            return self.visit_coreir_wire(module)
        if defn.coreir_name == "term":
            return True
        if defn.coreir_name == "undriven":
            mlir_type = hw.InOutType(module.results[0].type)
            wire = self._ctx.new_value(mlir_type)
            sym = self._ctx.parent.get_or_make_mapped_symbol(
                inst, name=f"{self._ctx.name}.{inst.name}", force=True)
            sv.WireOp(results=[wire], sym=sym)
            sv.ReadInOutOp(operands=[wire], results=module.results)
            return True
        if defn.coreir_name == "neg":
            zero = self.make_constant(type(inst.I), 0)
            comb.BaseOp(
                op_name="sub",
                operands=[zero, module.operands[0]],
                results=module.results)
            return True
        op_name = defn.coreir_name
        if op_name == "ashr":
            op_name = "shrs"
        if op_name == "lshr":
            op_name = "shru"
        comb.BaseOp(
            op_name=op_name,
            operands=module.operands,
            results=module.results)
        return True

    @wrap_with_not_implemented_error
    def visit_muxn(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert defn.coreir_name == "muxn"
        data = self._ctx.new_value(defn.I.data)
        sel = self._ctx.new_value(defn.I.sel)
        # NOTE(rsetaluri): This is a specialized code path for the case where
        # all tuples are flattened.
        if self._ctx.opts.flatten_all_tuples:
            hw.ArrayGetOp(
                operands=module.operands.copy(),
                results=module.results.copy())
            return True
        hw.StructExtractOp(
            field="data",
            operands=module.operands.copy(),
            results=[data])
        hw.StructExtractOp(
            field="sel",
            operands=module.operands.copy(),
            results=[sel])
        hw.ArrayGetOp(
            operands=[data, sel],
            results=module.results.copy())
        return True

    @wrap_with_not_implemented_error
    def visit_lutN(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert defn.coreir_name == "lutN"
        init = defn.coreir_configargs["init"]
        consts = [self.make_constant(Bit, b) for b in reversed(init)]
        mlir_type = hw.ArrayType((len(init),), builtin.IntegerType(1))
        array = self._ctx.new_value(mlir_type)
        hw.ArrayCreateOp(
            operands=consts,
            results=array)
        hw.ArrayGetOp(
            operands=[array, module.operands[0]],
            results=module.results)
        return True

    @wrap_with_not_implemented_error
    def visit_commonlib_primitive(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert defn.coreir_lib == "commonlib"
        if defn.coreir_name == "muxn":
            return self.visit_muxn(module)
        if defn.coreir_name == "lutN":
            return self.visit_lutN(module)

    @wrap_with_not_implemented_error
    def visit_array_get(self, module: ModuleWrapper) -> bool:
        inst_wrapper = module.module
        T = inst_wrapper.attrs["T"]
        size = T.N
        operands = module.operands
        index = inst_wrapper.attrs["index"]
        # NOTE(rsetaluri): This is "hacky" way to emit IR for ArrayGet(Array[1,
        # _], _) to work, since MLIR doesn't support i0 integer
        # constants. Instead, we form an Array[2, _] using a concat with a dummy
        # (const) element, and then perform ArrayGet on the Array[2, _] type
        # using an i1 constant.
        # TODO(rsetaluri): Figure out how to emit hw.array_get for
        # !hw.array_type<1x_> types.
        if size == 1:
            assert index == 0
            other = self.make_constant(T)
            concat_type = hw.ArrayType((2,), operands[0].type.T)
            concat = self._ctx.new_value(concat_type)
            hw.ArrayConcatOp(
                operands=[other, operands[0]],
                results=[concat])
            operands = [concat]
            size = 2
        num_sel_bits = clog2(size)
        index = self.make_constant(Bits[num_sel_bits], index)
        hw.ArrayGetOp(
            operands=(operands + [index]),
            results=module.results)
        return True

    @wrap_with_not_implemented_error
    def visit_array_slice(self, module: ModuleWrapper) -> bool:
        inst_wrapper = module.module
        T = inst_wrapper.attrs["T"]
        size = T.N
        operands = module.operands
        lo = inst_wrapper.attrs["lo"]
        num_sel_bits = clog2(size)
        lo = self.make_constant(Bits[num_sel_bits], lo)
        hw.ArraySliceOp(
            operands=(operands + [lo]),
            results=module.results)
        return True

    @wrap_with_not_implemented_error
    def visit_when(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert iswhen(defn)

        builder = defn._builder_

        input_to_index = {}
        output_to_index = {}

        def _visit(value, counter, value_to_index):
            # NOTE(leonardt): value may already be in value_to_index, in which
            # case we update it to a new index.  This is okay and it just means
            # that `value` occurs as more than one input (for example, it could
            # be used as a conditional driver for multiple values).  We could
            # optimize the logic to always share the same argument, but for now
            # we just use the "last" index.
            value_to_index[value] = next(counter)

        counter = itertools.count()
        for value in sort_by_value(builder.input_to_index):
            visit_magma_value_or_value_wrapper_by_direction(
                value,
                _assert_false,
                lambda v: _visit(v, counter, input_to_index),
                flatten_all_tuples=self._ctx.opts.flatten_all_tuples,
                inout_visitor=lambda v: _visit(v, counter, input_to_index),
            )
        counter = itertools.count()
        for value in sort_by_value(builder.output_to_index):
            visit_magma_value_or_value_wrapper_by_direction(
                value,
                lambda v: _visit(v, counter, output_to_index),
                _assert_false,
                flatten_all_tuples=self._ctx.opts.flatten_all_tuples,
                inout_visitor=lambda v: _visit(v, counter, output_to_index),
            )

        wires = [
            self._ctx.new_value(hw.InOutType(result.type))
            for result in module.results
        ]
        for result, wire in zip(module.results, wires):
            sv.RegOp(results=[wire])
            sv.ReadInOutOp(operands=[wire], results=[result])

        def _collect_visited(value):
            fields = []
            visit_magma_value_or_value_wrapper_by_direction(
                value, fields.append, fields.append,
                flatten_all_tuples=self._ctx.opts.flatten_all_tuples,
                inout_visitor=fields.append,
            )
            return fields

        def _make_assignments(connections):
            for drivee, driver in connections:
                elts = zip(*map(_collect_visited, (drivee, driver)))
                for drivee_elt, driver_elt in elts:
                    operand = module.operands[input_to_index[driver_elt]]
                    wire = wires[output_to_index[drivee_elt]]
                    sv.BPAssignOp(operands=[wire, operand])

        def _process_when_block(block):
            connections = (
                (conditional_wire.drivee, conditional_wire.driver)
                for conditional_wire in block.conditional_wires()
            )
            if block.condition is None:
                _make_assignments(connections)
                for child in block.children():
                    _process_when_block(child)
                return
            cond = module.operands[input_to_index[block.condition]]
            if_op = sv.IfOp(operands=[cond])
            with push_block(if_op.then_block):
                _make_assignments(connections)
                for child in block.children():
                    _process_when_block(child)
            curr_sibling = if_op
            with contextlib.ExitStack() as stack:
                sibling_blocks = list(block.elsewhen_blocks())
                if block.otherwise_block is not None:
                    sibling_blocks.append(block.otherwise_block)
                for sibling_block in sibling_blocks:
                    stack.enter_context(push_block(curr_sibling.else_block))
                    curr_sibling = _process_when_block(sibling_block)
            return if_op

        with push_block(sv.AlwaysCombOp().body_block):
            _make_assignments(builder.default_drivers.items())
            _process_when_block(builder.block)

        return True

    @wrap_with_not_implemented_error
    def visit_primitive(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert isprimitive(defn)
        if defn.coreir_lib == "coreir" or defn.coreir_lib == "corebit":
            return self.visit_coreir_primitive(module)
        if defn.coreir_lib == "commonlib":
            return self.visit_commonlib_primitive(module)
        if isinstance(defn, InlineVerilogExpression):
            assert len(module.operands) == 0
            assert len(module.results) > 0
            sv.VerbatimExprOp(
                operands=list(),
                results=module.results,
                expr=defn.expr,
            )
            return True
        if iswhen(defn):
            return self.visit_when(module)

    @wrap_with_not_implemented_error
    def visit_magma_wire(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert isinstance(defn, Wire) and not defn.flattened
        return self.visit_coreir_wire(module)

    @wrap_with_not_implemented_error
    def visit_magma_mux(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert isinstance(defn, Mux)
        # NOTE(rsetaluri): This is a round-about way to get the height while
        # magma.Mux does not store those parameters. That should be fixed in
        # magma/primitives/mux.py.
        height = len(list(filter(
            lambda p: "I" in p.name.name, defn.interface.outputs())))
        data, select = module.operands[:-1], module.operands[-1]
        # NOTE(rsetaluri): This is a specialized code path for the case where
        # all tuples are flattened.
        if self._ctx.opts.flatten_all_tuples and len(data) != height:
            assert len(data) % height == 0
            stride = len(data) // height
            assert len(module.results) == stride
            for i in range(stride):
                inputs = list(reversed(data[i::stride]))
                make_mux(self._ctx, inputs, select, module.results[i])
            return True
        # NOTE(rsetaluri): Reversing data needs to be done *after* we check for
        # tuple flattening. Otherwise the tuple field ordering gets reversed as
        # well.
        data = list(reversed(data))
        make_mux(self._ctx, data, select, module.results[0])
        return True

    @wrap_with_not_implemented_error
    def visit_magma_register(self, module: ModuleWrapper) -> bool:
        # NOTE(rsetaluri): Refactoring this compilation into the make_register
        # and get_register_data_and_init functions is necessary to support
        # flatten_all_tuples=True. make_register works for arbitrary types of
        # inputs, so we can either invoke it once for the top-level type or call
        # it on all flattened parts. get_register_data_and_init associates the
        # operands with init values (and magma type), so that flattened and
        # unflattened tuples can be treated uniformly.

        def make_register(T, data, init, result):
            reg = self._ctx.new_value(hw.InOutType(data.type))
            sv.RegOp(name=inst.name, results=[reg])
            always = sv.AlwaysFFOp(operands=always_operands, **attrs)
            const = self.make_constant(T, init)
            with push_block(always.body_block):
                ctx = contextlib.nullcontext()
                if has_enable:
                    ctx = push_block(sv.IfOp(operands=[enable]).then_block)
                with ctx:
                    sv.PAssignOp(operands=[reg, data])
            if has_reset:
                with push_block(always.reset_block):
                    sv.PAssignOp(operands=[reg, const])
            if not self._ctx.opts.disable_initial_blocks:
                with push_block(sv.InitialOp()):
                    sv.BPAssignOp(operands=[reg, const])
            sv.ReadInOutOp(operands=[reg], results=[result])

        inst = module.module
        defn = type(inst)
        clock_edge = "posedge"
        has_reset = defn.reset_type is not None
        # NOTE(resetaluri): This is a hack until
        # magma/primitives/register.py:Register is updated to store this
        # generator parameter directly.
        has_enable = "CE" in defn.interface.ports
        operands = module.operands.copy()
        T, data, init = get_register_data_and_init(
            self._ctx, defn.I, operands, defn.init)
        assert len(data) == len(init)
        assert len(data) == len(T)
        assert len(data) == len(module.results)
        if has_enable:
            enable = operands[0]
            clk = operands[1]
        else:
            clk = operands[0]
        always_operands = [clk]
        attrs = dict(clock_edge=clock_edge)
        if has_reset:
            reset = operands[-1]
            always_operands.append(reset)
            reset_type, reset_edge = parse_reset_type(defn.reset_type)
            attrs.update(dict(reset_type=reset_type, reset_edge=reset_edge))
        for T_i, data_i, init_i, result_i in zip(T, data, init, module.results):
            make_register(T_i, data_i, init_i, result_i)
        return True

    @wrap_with_not_implemented_error
    def visit_magma_xmr_sink(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert isinstance(defn, XMRSink)
        xmr = defn.value
        try:
            self._ctx.parent.xmr_paths[xmr]
        except KeyError:
            pass
        else:
            return True
        paths = get_xmr_paths(self._ctx, xmr)
        assert len(paths) == len(module.operands)
        self._ctx.parent.xmr_paths[xmr] = paths
        return True

    @wrap_with_not_implemented_error
    def visit_magma_xmr_source(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        assert isinstance(defn, XMRSource)
        xmr = defn.value
        paths = self._ctx.parent.xmr_paths[xmr]
        assert len(paths) == len(module.results)
        base = defn.value.parent.path()
        for result, path in zip(module.results, paths):
            in_out = self._ctx.new_value(hw.InOutType(result.type))
            path = base + path
            sv.XMROp(is_rooted=False, path=path, results=[in_out])
            sv.ReadInOutOp(operands=[in_out], results=[result])
        return True

    @wrap_with_not_implemented_error
    def visit_inline_verilog(self, module: ModuleWrapper) -> bool:
        inst = module.module
        defn = type(inst)
        inline_verilog_strs = defn.inline_verilog_strs
        assert inline_verilog_strs
        for string, references in inline_verilog_strs:
            # NOTE(rsetaluri): We assume that the order of the references
            # matches the order of the ports to the encapsulating inline verilog
            # module (which is safe as of phanrahan/magma:d3e8c95).
            replacement_map = {
                k: f"{{{{{i}}}}}" for i, k in enumerate(references)
            }
            # NOTE(rsetaluri): We need to traverse the replacements in order of
            # decreasing key-length (i.e. strlen) in order to ensure that we
            # don't replace e.g. "key10" with "{repl}0" where both "key1" and
            # "key10" are keys.
            replacements = reversed(sorted(
                replacement_map.items(), key=lambda kv: len(kv[1])))
            for k, v in replacements:
                k = "{" + k + "}"
                string = string.replace(k, v)
            sv.VerbatimOp(operands=module.operands, string=string)
        return True

    @wrap_with_not_implemented_error
    def visit_instance(self, module: ModuleWrapper) -> bool:
        inst = module.module
        assert isinstance(inst, AnonymousCircuitType)
        defn = type(inst)
        elaborate_magma_registers = self._ctx.opts.elaborate_magma_registers
        if isinstance(defn, Wire) and not defn.flattened:
            return self.visit_magma_wire(module)
        if isinstance(defn, Mux):
            return self.visit_magma_mux(module)
        if isinstance(defn, Register) and not elaborate_magma_registers:
            return self.visit_magma_register(module)
        if isinstance(defn, XMRSink):
            return self.visit_magma_xmr_sink(module)
        if isinstance(defn, XMRSource):
            return self.visit_magma_xmr_source(module)
        if getattr(defn, "inline_verilog_strs", []):
            return self.visit_inline_verilog(module)
        if isprimitive(defn):
            return self.visit_primitive(module)
        module_type = self._ctx.parent.get_hardware_module(defn).hw_module
        parameters = filter_by_key(
            lambda key: key not in ["name", "locl"],
            inst.kwargs
        )
        compile_guard = get_compile_guard_data(inst)
        sym = None
        attrs = {}
        if is_bound_instance(inst):
            sym = self._ctx.parent.get_or_make_mapped_symbol(
                inst, name=f"{inst.defn.name}.{inst.name}", force=True
            )
            attrs["doNotPrint"] = 1
        make_hw_instance_op(
            name=inst.name,
            module=module_type,
            operands=module.operands,
            results=module.results,
            sym=sym,
            parameters=parameters,
            compile_guard=compile_guard,
            attrs=attrs)
        return True

    @wrap_with_not_implemented_error
    def visit_instance_wrapper(self, module: ModuleWrapper) -> bool:
        inst_wrapper = module.module
        assert isinstance(inst_wrapper, MagmaInstanceWrapper)
        if inst_wrapper.name.startswith("magma_array_get_op_"):
            T = inst_wrapper.attrs["T"]
            if isinstance(T, BitsMeta) or issubclass(T.T, Bit):
                comb.ExtractOp(
                    operands=module.operands,
                    results=module.results,
                    lo=inst_wrapper.attrs["index"])
                return True
            return self.visit_array_get(module)
        if inst_wrapper.name.startswith("magma_array_slice_op_"):
            T = inst_wrapper.attrs["T"]
            if isinstance(T, BitsMeta) or issubclass(T.T, Bit):
                comb.ExtractOp(
                    operands=module.operands,
                    results=module.results,
                    lo=inst_wrapper.attrs["lo"])
                return True
            return self.visit_array_slice(module)
        if inst_wrapper.name.startswith("magma_array_create_op"):
            T = inst_wrapper.attrs["T"]
            if isinstance(T, BitsMeta) or issubclass(T.T, Bit):
                comb.ConcatOp(
                    operands=list(reversed(module.operands)),
                    results=module.results)
                return True
            hw.ArrayCreateOp(
                operands=list(reversed(module.operands)),
                results=module.results)
            return True
        if inst_wrapper.name.startswith("magma_tuple_get_op"):
            index = inst_wrapper.attrs["index"]
            hw.StructExtractOp(
                field=index,
                operands=module.operands,
                results=module.results)
            return True
        if inst_wrapper.name.startswith("magma_tuple_create_op"):
            hw.StructCreateOp(
                operands=module.operands,
                results=module.results)
            return True
        is_const = (
            inst_wrapper.name.startswith("magma_bit_constant_op") or
            inst_wrapper.name.startswith("magma_bits_constant_op"))
        if is_const:
            value = inst_wrapper.attrs["value"]
            hw.ConstantOp(value=int(value), results=module.results)
            return True

    @wrap_with_not_implemented_error
    def visit_module(self, module: ModuleWrapper) -> bool:
        if isinstance(module.module, DefineCircuitKind):
            return True
        if isinstance(module.module, AnonymousCircuitType):
            return self.visit_instance(module)
        if isinstance(module.module, MagmaInstanceWrapper):
            return self.visit_instance_wrapper(module)

    def _process_magma_module(self, magma_module: MagmaModuleLike):
        if not self._graph.has_node(magma_module):
            return
        for predecessor in self._graph.predecessors(magma_module):
            if predecessor in self._visited:
                continue
            self.visit(predecessor)
        for src, _, data in self._graph.in_edges(magma_module, data=True):
            info = data["info"]
            src_port, dst_port = info["src"], info["dst"]
            src_value = self._ctx.get_or_make_mapped_value(src_port)
            self._ctx.set_mapped_value(dst_port, src_value)

    def visit(self, module: MagmaModuleLike):
        if module in self._visited:
            raise ModuleVisitor.RevisitedModuleError(module)
        self._visited.add(module)
        self._process_magma_module(module)
        self.visit_module(ModuleWrapper.make(module, self._ctx))
        instances = getattr(module, "instances", [])
        for inst in instances:
            if inst in self._visited:
                continue
            self.visit(inst)


def treat_as_primitive(
        defn_or_decl: CircuitKind,
        ctx: 'HardwareModule'
) -> bool:
    # NOTE(rsetaluri): This is a round-about way to mark new types as
    # primitives. These definitions should actually be marked as primitives.
    elaborate_magma_registers = ctx.opts.elaborate_magma_registers
    if isprimitive(defn_or_decl):
        return True
    if isinstance(defn_or_decl, Mux):
        return True
    if isinstance(defn_or_decl, Register) and not elaborate_magma_registers:
        return True
    if getattr(defn_or_decl, "inline_verilog_strs", []):
        return True
    return False


def treat_as_definition(defn_or_decl: CircuitKind) -> bool:
    if not isdefinition(defn_or_decl):
        return False
    if getattr(defn_or_decl, "verilog", ""):
        return False
    if getattr(defn_or_decl, "verilogFile", ""):
        return False
    return True


class BindProcessorInterface(abc.ABC):
    def __init__(self, ctx: 'HardwareModule', defn: CircuitKind):
        self._ctx = ctx
        self._defn = defn

    @abc.abstractmethod
    def preprocess(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def process(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def postprocess(self):
        raise NotImplementedError()


class NativeBindProcessor(BindProcessorInterface):
    def preprocess(self):
        for bind_module in self._defn.bind_modules:
            # TODO(rsetaluri): Here we should check if @bind_module has already
            # been compiled, in the case that the bound module is either used
            # multiple times or is also a "normal" module that was compiled
            # elsewhere. Currently, the `set_hardware_module` call will raise an
            # error if @bind_module has been compiled already.
            hardware_module = self._ctx.parent.new_hardware_module(bind_module)
            hardware_module.compile()
            assert hardware_module.hw_module is not None
            self._ctx.parent.set_hardware_module(
                bind_module, hardware_module.hw_module)

    @wrap_with_not_implemented_error
    def _resolve_arg(self, arg) -> MlirValue:
        if isinstance(arg, Type):
            return self._ctx.get_mapped_value(arg)
        if isinstance(arg, PortView):
            return resolve_xmr(self._ctx, arg)

    def process(self):
        self._syms = []
        for bind_module, (args, _) in self._defn.bind_modules.items():
            operands = []
            for port in self._defn.interface.ports.values():
                port = get_magma_value(port)
                visit_magma_value_or_value_wrapper_by_direction(
                    port,
                    lambda p: operands.append(self._ctx.get_mapped_value(p)),
                    lambda p: operands.append(self._ctx.get_mapped_value(p)),
                    flatten_all_tuples=self._ctx.opts.flatten_all_tuples,
                )
            operands += list(map(self._resolve_arg, args))
            inst_name = f"{bind_module.name}_inst"
            sym = self._ctx.parent.get_or_make_mapped_symbol(
                (self._defn, bind_module),
                name=f"{self._defn.name}.{inst_name}",
                force=True)
            module = self._ctx.parent.get_hardware_module(bind_module)
            inst = hw.InstanceOp(
                name=inst_name,
                module=module,
                operands=operands,
                results=[],
                sym=sym)
            inst.attr_dict["doNotPrint"] = 1
            self._syms.append(sym)

    def postprocess(self):
        defn_sym = self._ctx.parent.get_mapped_symbol(self._defn)
        for sym in self._syms:
            instance = hw.InnerRefAttr(defn_sym, sym)
            sv.BindOp(instance=instance)
        bound_instances = list(filter(is_bound_instance, self._defn.instances))
        for bound_instance in bound_instances:
            inst_sym = self._ctx.parent.get_mapped_symbol(bound_instance)
            ref = hw.InnerRefAttr(defn_sym, inst_sym)
            sv.BindOp(instance=ref)
        if self._syms or bound_instances:
            bind_filename = f"{self._ctx.opts.basename}.sv"
            self._ctx.parent.add_bind_file(bind_filename)


class CoreIRBindProcessor(BindProcessorInterface):
    def preprocess(self):
        return

    def process(self):
        for name, content in self._defn.compiled_bind_modules.items():
            path = pathlib.Path(self._ctx.opts.basename).parent
            filename = path / f"{name}.sv"
            with open(filename, "w") as f:
                f.write(content)
            self._ctx.parent.add_bind_file(filename.name)

    def postprocess(self):
        return


def _make_bind_processor(ctx: 'HardwareModule', defn: CircuitKind):
    if ctx.opts.use_native_bind_processor:
        return NativeBindProcessor(ctx, defn)
    return CoreIRBindProcessor(ctx, defn)


def _visit_linked_module(
        ctx: 'HardwareModule', decl: CircuitKind, op: hw.ModuleOp):
    # In order to emit a linked module, we perform the following steps:
    #   (1) Declare a wire for each output of the module.
    #   (2) For each linked module, first enter an `ifdef block.
    #       (a) Instantiate the linked target.
    #       (b) Assign the wires with the outputs of the instance.
    #       (c) Enter the `else of the `ifdef block.
    #   (3) If we have a default link target, perform (2a-b).
    module = ModuleWrapper.make(decl, ctx)
    wires = {
        output: ctx.new_value(hw.InOutType(output.type))
        for output in module.operands
    }

    def _process_target(defn):
        hw_module = ctx.parent.get_hardware_module(defn).hw_module
        results = [ctx.new_value(output.type) for output in module.operands]
        make_hw_instance_op(
            name=f"{defn.name}_inst",
            module=hw_module,
            operands=module.results,
            results=results,
        )
        for result, wire in zip(results, wires.values()):
            sv.AssignOp(operands=[wire, result])

    with push_block(op):
        for output, wire in wires.items():
            sv.WireOp(results=[wire])
            sv.ReadInOutOp(operands=[wire], results=[output])
        with contextlib.ExitStack() as stack:
            for key, linked_module in get_linked_modules(decl).items():
                if_def = sv.IfDefOp(key)
                stack.enter_context(push_block(if_def.then_block))
                _process_target(linked_module)
                stack.enter_context(push_block(if_def.else_block))
            if has_default_linked_module(decl):
                default_linked_module = get_default_linked_module(decl)
                _process_target(default_linked_module)
        hw.OutputOp(operands=module.operands)
    return op


class HardwareModule:
    def __init__(
            self, magma_defn_or_decl: CircuitKind,
            parent: weakref.ReferenceType, opts: CompileToMlirOpts):
        self._magma_defn_or_decl = magma_defn_or_decl
        self._parent = parent
        self._opts = opts
        self._hw_module = None
        self._name_gen = ScopedNameGenerator(disallow_duplicates=False)
        self._value_map = {}

    @property
    def magma_defn_or_decl(self) -> CircuitKind:
        return self._magma_defn_or_decl

    @property
    def parent(self):
        return self._parent()

    @property
    def opts(self) -> CompileToMlirOpts:
        return self._opts

    @property
    def hw_module(self) -> hw.ModuleOpBase:
        return self._hw_module

    @property
    def name(self) -> str:
        return self._magma_defn_or_decl.name

    def get_mapped_value(self, port: Type) -> MlirValue:
        return self._value_map[port]

    def get_or_make_mapped_value(self, port: Type, **kwargs) -> MlirValue:
        try:
            return self._value_map[port]
        except KeyError:
            pass
        self._value_map[port] = value = self.new_value(port, **kwargs)
        return value

    def set_mapped_value(self, port: Type, value: MlirValue):
        if port in self._value_map:
            raise ValueError(f"Port {port} already mapped")
        self._value_map[port] = value

    def new_value(
            self, value_or_type: Union[Type, Kind, MlirType],
            **kwargs) -> MlirValue:
        if isinstance(value_or_type, Type):
            mlir_type = magma_type_to_mlir_type(type(value_or_type))
        elif isinstance(value_or_type, Kind):
            mlir_type = magma_type_to_mlir_type(value_or_type)
        elif isinstance(value_or_type, MlirType):
            mlir_type = value_or_type
        elif isinstance(value_or_type, MagmaValueWrapper):
            mlir_type = magma_type_to_mlir_type(value_or_type.T)
        else:
            raise TypeError(value_or_type)
        name = self._name_gen(**kwargs)
        return MlirValue(mlir_type, name)

    def compile(self):
        self._hw_module = self._compile()
        if self._hw_module is None:
            return
        self._add_module_parameters(self._hw_module)
        _maybe_set_location_info(
            self._hw_module,
            self._magma_defn_or_decl.debug_info
        )

    def _compile(self) -> hw.ModuleOpBase:
        if treat_as_primitive(self._magma_defn_or_decl, self):
            return

        def new_values(fn, ports):
            namer = magma_value_or_type_to_string
            return [
                fn(port, name=namer(port), force=True)
                for port in map(get_magma_value, ports)
            ]

        i, o = [], []
        for port in self._magma_defn_or_decl.interface.ports.values():
            port = get_magma_value(port)
            visit_magma_value_or_value_wrapper_by_direction(
                port, i.append, o.append,
                flatten_all_tuples=self._opts.flatten_all_tuples,
            )
        inputs = new_values(self.get_or_make_mapped_value, o)
        named_outputs = new_values(self.new_value, i)
        defn_or_decl_output_name = _get_defn_or_decl_output_name(
            self, self._magma_defn_or_decl)
        name = self.parent.get_or_make_mapped_symbol(
            self._magma_defn_or_decl,
            name=defn_or_decl_output_name, force=True)
        if has_any_linked_modules(self._magma_defn_or_decl):
            op = hw.ModuleOp(
                name=name,
                operands=inputs,
                results=named_outputs,
            )
            return _visit_linked_module(self, self._magma_defn_or_decl, op)
        if not treat_as_definition(self._magma_defn_or_decl):
            return hw.ModuleExternOp(
                name=name,
                operands=inputs,
                results=named_outputs)
        bind_processor = _make_bind_processor(self, self._magma_defn_or_decl)
        bind_processor.preprocess()
        op = hw.ModuleOp(
            name=name,
            operands=inputs,
            results=named_outputs)
        build_magma_graph_opts = BuildMagmaGrahOpts(
            self._opts.flatten_all_tuples)
        graph = build_magma_graph(
            self._magma_defn_or_decl, build_magma_graph_opts)
        visitor = ModuleVisitor(graph, self)
        with push_block(op):
            try:
                visitor.visit(self._magma_defn_or_decl)
            except ModuleVisitor.VisitError:
                raise MlirCompilerInternalError(visitor)
            bind_processor.process()
            output_values = new_values(self.get_or_make_mapped_value, i)
            if named_outputs:
                hw.OutputOp(operands=output_values)
        bind_processor.postprocess()
        return op

    def _add_module_parameters(self, hw_module: hw.ModuleOpBase):
        defn_or_decl = self._magma_defn_or_decl
        try:
            param_types = defn_or_decl.coreir_config_param_types
        except AttributeError:
            return
        for name, type in param_types.items():
            type = python_type_to_mlir_type(type)
            param = hw.ParamDeclAttr(name, type)
            hw_module.parameters.append(param)



from ..array import Array
from ..bit import Bit, Digital
from ..compiler import make_compiler
from ..passes.debug_name import DebugNamePass
from ..passes.clock import WireClockPass
from collections import OrderedDict
from ..logging import root_logger
from graphviz import Digraph
import re


_logger = root_logger()


# Initially based on the firrtl backend.
def escape(s):
    # Escape special characters for graphviz labels.
    return re.sub(r'([<>])', r'\\\1', s)

def get_type(port):
    return str(type(port))

def get_name(dot, port):
    if isinstance(port, Digital) and port.const():
        return "1" if port is type(port).VCC else "0"

    if isinstance(port, Array):
        if not port.iswhole():
            # the sequence of values is concantenated
            port = [get_name(dot, i) for i in port.ts]
            port.reverse()
            if len(port) == 1:  # FIXME: Hack to make single length bit arrays work
                return port[0]
            name = '[' + ','.join(port) + ']'
            dot.node(name, '{{' + '|'.join(port) + '}|}')
            for sub in port:
                dot.edge(sub, name + ':' + sub)
            return name
    assert not port.anon()
    return port.name.qualifiedname(sep="_")

def compileinstance(dot, instance):
    instance_backend_name = str(instance.name)
    instance_cls_name = str(instance.__class__.__name__)

    inputs = []
    outputs = []

    for name, port in instance.interface.ports.items():
        if port.is_input():
            inputs.append('<{0}> {0}'.format(str(name)))
            value = port.value()
            if value is None:
                _logger.warning(f'Input {port.debug_name} not connected to an output')
                continue
        else:
            outputs.append('<{0}> {0}'.format(str(name)))
            value = port
        # if isinstance(value, Array):
        #     for index, subport in enumerate(value.ts):
        #         s += "{}.{}[{}] <= {}\n".format(instance_backend_name, name, index, get_name(subport))
        # else:
        value_name = get_name(dot, value)
        if port.is_input():
            # s += "{}.{} <= {}\n".format(instance_backend_name, name, value_name)
            dot.edge(value_name, '{}:{}'.format(instance_backend_name, name))
        else:
            # s += "{} <= {}.{}\n".format(value_name, instance_backend_name, name)
            dot.edge('{}:{}'.format(instance_backend_name, name), value_name)

    instance_label = '{' + '|'.join(inputs) + '}|'
    if instance.decl is not None:
        instance_label += '{} ({})\\n{}'.format(instance.decl.varname, instance_backend_name, instance_cls_name)
    else:
        instance_label += '{}\\n{}'.format(instance_backend_name, instance_cls_name)

    instance_label += '|{' + '|'.join(outputs) + '}'
    dot.node(instance_backend_name, '{' + instance_label + '}')

def compiledefinition(dot, cls):
    # Each definition maps to an entire self-contained graphviz digraph.

    # for now only allow Bit or Array(n, Bit)
    for name, port in cls.interface.ports.items():
        if isinstance(port, Array):
            if not issubclass(port.T, Bit):
                print('Error: Argument', port, 'must be a an Array(n,Bit)')

    for name, port in cls.interface.ports.items():
        # TODO: Check whether port.is_input() or .is_output().
        dot.node(name,
                 escape('{}\\n{}'.format(name, get_type(port))),
                 shape='ellipse')

    # declare a wire for each instance output
    for instance in cls.instances:
        for port in instance.interface.ports.values():
            if port.is_output():
                dot.node(get_name(dot, port),
                         escape('{}\\n{}'.format(get_name(dot, port), get_type(port))),
                         shape='none')

    # Emit the graph node (with ports) for each instance.
    for instance in cls.instances:
        compileinstance(dot, instance)

    # Assign to module output arguments.
    for input in cls.interface.inputs():
        output = input.value()
        if output is not None:
            iname = get_name(dot, input)
            oname = get_name(dot, output)
            dot.edge(oname, iname)

def find(circuit, defn):
    name = circuit.name
    if not hasattr(circuit, "instances"):
        return defn
    for i in circuit.instances:
        find(type(i), defn)
    if name not in defn:
        defn[name] = circuit
    return defn

def dots(main):
    defn = find(main,OrderedDict())

    dots = []
    for k, v in defn.items():
         # print('compiling', k, v)
         dot = Digraph(k,
                       graph_attr={'label': str(v), 'rankdir': 'LR'},
                       node_attr={'shape': 'record'})
         compiledefinition(dot, v)
         dots.append(dot)

    return dots

def to_html(main):
    DebugNamePass(main).run()
    WireClockPass(main).run()
    return "\n".join([dot._repr_svg_() for dot in dots(main)])

def compile(main):
    DebugNamePass(main).run()
    WireClockPass(main).run()
    return "\n".join([str(dot) for dot in dots(main)]) + '\n'

DotCompiler = make_compiler("dot", compile)



import sys
from abc import abstractmethod
if sys.version_info < (3, 4):
    import abc
    ABC = abc.ABCMeta('ABC', (object,), {})
else:
    from abc import ABC
from collections import namedtuple
from itertools import product
from .simulator import CircuitSimulator, ExecutionState
from ..transforms import flatten
from ..circuit import *
from ..scope import *
from ..bit import Bit, Digital
from ..array import Array
from ..tuple import Tuple
from ..bits import SInt, Bits, UInt
from hwtypes import BitVector
import hwtypes
from ..bitutils import seq2int
from ..clock import Clock
from ..passes.clock import WireClockPass

__all__ = ['PythonSimulator']

ExecutionOrder = namedtuple('ExecutionOrder', ['stateful', 'combinational'])


class PythonSimulatorException(Exception):
    pass


class SimPrimitive:
    def __init__(self, primitive, value_store):
        if primitive.simulate is None:
            raise ValueError("Cannot simulate {} of type {} because it does not have a Python simulate method".format(primitive, type(primitive)))
        self.primitive = primitive
        self.inputs = []
        self.outputs = []
        self.value_store = value_store
        self.state_store = {}

        for bit in self.primitive.interface.ports.values():
            if not isinstance(bit, Array):
                bit = [bit]
            for b in bit:
                if b.is_input():
                    self.inputs.append(b)
                else:
                    self.outputs.append(b)

    def stateful(self):
        return self.primitive.stateful

    def inputs_satisfied(self):
        for i in self.inputs:
            if not self.value_store.value_initialized(i):
                return False

        return True

    def initialize_outputs(self):
        # Initializes all outputs to False, should perhaps
        # initialize based on starting inputs?
        for o in self.outputs:
            self.value_store.set_value(o, False)

    def simulate(self):
        self.primitive.simulate(self.value_store, self.state_store)

class WatchPoint:
    idx = 0
    def __init__(self, bit, scope, simulator, value):
        self.bit = bit
        self.scope = scope
        self.simulator = simulator
        self.value = value
        self.old_val = self.simulator.get_value(self.bit, self.scope)

        WatchPoint.idx += 1
        self.idx = WatchPoint.idx

    def was_triggered(self):
        new_val = self.simulator.get_value(self.bit, self.scope)
        triggered = new_val != self.old_val
        if self.value:
            if self.value != new_val:
                triggered = False
        self.old_val = new_val

        return triggered

class ValueStore:
    def __init__(self):
        self.value_map = {}

    def value_initialized(self, bit):
        if isinstance(bit, (Array, Tuple)):
            for b in bit:
                if not self.value_initialized(b):
                    return False

            return True

        if bit.is_input():
            bit = bit.value()

        if bit.const():
            return True

        return bit in self.value_map

    def get_value(self, bit):
        if isinstance(bit, Array):
            value = [self.get_value(b) for b in bit]
            if isinstance(bit, SInt):
                return BitVector[len(bit)](value).as_sint()
            elif isinstance(bit, UInt):
                return BitVector[len(bit)](value).as_uint()
            elif isinstance(bit, Bits):
                return BitVector[len(bit)](value)
            return value

        if bit.is_input():
            bit = bit.value()

        if bit.const():
            return bool(bit)

        return self.value_map[bit]

    def set_value(self, bit, newval):
        if not bit.is_output():
            raise TypeError("Can only call set value on an input")

        if isinstance(bit, Array):
            if isinstance(newval, BitVector):
                newval = newval.as_bool_list()
            elif isinstance(newval, Bits):
                if not newval.const():
                    raise ValueError("Calling set_value with a Bits only works with a constant")
                newval = newval.bits()
            elif isinstance(bit, Bits) and isinstance(newval, int):
                if not isinstance(bit, SInt) and newval < 0:
                    raise ValueError(f"Can only set {bit} of type {type(bit)} with positive integer, not {newval}")
                newval = BitVector[len(bit)](newval).as_bool_list()
            elif not isinstance(newval, list):
                raise TypeError(f"Calling set_value with {bit} of type {type(bit)} only works with a list of values or a BitVector")

            for b,v in zip(bit, newval):
                self.set_value(b, v)
            return

        if isinstance(newval, int) and newval in {0, 1}:
            newval = bool(newval)
        if isinstance(newval, hwtypes.Bit):
            newval = bool(newval)
        if not isinstance(newval, bool):
            raise TypeError(f"Can only set Bit {bit} with a boolean value or 0 or 1, not {newval} (type={type(newval)})")

        self.value_map[bit] = newval

class PythonSimulator(CircuitSimulator):
    def __setup_primitives(self):
        wrapped = []
        for primitive in self.circuit.instances:
            wrapped.append(SimPrimitive(primitive, self.value_store))

        return wrapped

    def initialize(self, bit):
        if isinstance(bit, (Array, Tuple)):
            for b in bit:
                self.initialize(b)
        else:
            if bit.is_output():
                self.circuit_inputs.append(bit)
                self.value_store.set_value(bit, False)
            else:
                self.circuit_outputs.append(bit)

    def __setup_circuit(self, clock):
        if clock is not None:
            clock = self.txfm.get_new_bit(clock, self.default_scope)
        self.clock = clock

        self.circuit_inputs = []
        self.circuit_outputs = []
        for name, bit in self.circuit.interface.ports.items():
            self.initialize(bit)

    def __outputs_initialized(self):
        for bit in self.circuit_outputs:
            assert bit.is_input()
            if not self.value_store.value_initialized(bit):
                return False

        return True

    def __sort_state_primitives(self, state_primitives):
        """
        State primitives should be sorted in reversed topological order.
        This ensures that the simulation order of the stateful elements
        is correct.
        Intuition:
            If the output value of a state element `x` feeds into the input of
            another state element `y`,
            `y` should perform it's simulation before `x` because it will use
            the value of the signal on the previous clock cycle.
        """
        sorted_state_primitives = []
        for primitive_1 in state_primitives:
            inserted = False
            for index, primitive_2 in enumerate(sorted_state_primitives):
                for output in primitive_1.inputs:
                    if any(output.value() is x for x in primitive_2.inputs):
                        sorted_state_primitives.insert(index, primitive_1)
                        inserted = True
                        break
                if inserted:
                    break
            if not inserted:
                sorted_state_primitives.append(primitive_1)
        sorted_state_primitives.reverse()
        return sorted_state_primitives

    def __get_ordered_primitives(self, unordered_primitives):
        state_primitives = []
        after_state = []

        state_primitives = []
        for primitive in unordered_primitives:
            if primitive.stateful():
                primitive.initialize_outputs()
                state_primitives.append(primitive)

        sorted_state_primitives = self.__sort_state_primitives(state_primitives)

        unordered_primitives[:] = [u for u in unordered_primitives if not u.stateful()]

        combinational = []
        while len(unordered_primitives) > 0:
            found = False
            for primitive in unordered_primitives:
                if primitive.inputs_satisfied():
                    primitive.initialize_outputs()
                    combinational.append(primitive)
                    unordered_primitives.remove(primitive)
                    found = True
                    break
            assert found, "Some circuits have unsatisfied inputs"

        return ExecutionOrder(stateful=sorted_state_primitives, combinational=combinational)

    def __step(self):
        if self.clock is None:
            raise PythonSimulatorException("Cannot step a simulated circuit "
                    "without a clock, did you pass a clock during "
                    "initialization?")
        cur_clock_val = self.value_store.get_value(self.clock)
        self.value_store.set_value(self.clock, not cur_clock_val)

    def __init__(self, main_circuit, clock=None):
        if isinstance(main_circuit, CircuitType):
            raise ValueError("PythonSimulator must be called with a Circuit definition, not an instance")
        if clock is not None and not isinstance(clock, Clock):
            raise ValueError("clock must be a Clock or None")
        WireClockPass(main_circuit).run()
        self.main_circuit = main_circuit
        self.txfm = flatten(main_circuit)
        self.circuit = self.txfm.circuit
        self.value_store = ValueStore()
        self.default_scope = Scope()
        self.__setup_circuit(clock)
        self.watchpoints = []

        primitives = self.__setup_primitives()

        self.execution_order = self.__get_ordered_primitives(primitives)

        assert self.__outputs_initialized(), "All circuit outputs not initialized."

    def get_capabilities(self):
        return []

    def get_value(self, bit, scope=None):
        if scope is None:
            scope = self.default_scope
        newbit = self.txfm.get_new_bit(bit, scope)
        if newbit is None:
            return None

        try:
            return self.value_store.get_value(newbit)
        except KeyError:
            return None

    def set_value(self, bit, newval, scope=None):
        if scope is None:
            scope = self.default_scope
        newbit = self.txfm.get_new_bit(bit, scope)
        if not self.is_circuit_input(newbit):
            message = "Only setting main's inputs is supported (Trying to set: {})".format(bit)
            raise PythonSimulatorException(message)
        else:
            self.value_store.set_value(newbit, newval)

    def is_circuit_input(self, value):
        """
        Checks if `value` is in `self.circuit_inputs`.
        If `value` is an `Array`, it recursively checks the elements
        """
        if isinstance(value, Digital):
            return any(value is x for x in self.circuit_inputs)
        elif isinstance(value, Array):
            return all(self.is_circuit_input(elem) for elem in value)
        else:
            raise NotImplementedError(type(value))

    def advance(self, n=1):
        cycles = 0
        for i in range(0, n):
            self.__step()
            state = self.evaluate()

            if not state.clock:
                cycles += 1
            if state.triggered_points:
                return ExecutionState(triggered_points=state.triggered_points, clock=state.clock, cycles=cycles)

        return ExecutionState(triggered_points=[], clock=self.get_clock_value(), cycles=cycles)

    def evaluate(self, no_update=False):
        for primitive in self.execution_order.stateful:
            primitive.simulate()
        for primitive in self.execution_order.combinational:
            primitive.simulate()

        triggered = []
        for watch in self.watchpoints:
            if watch.was_triggered():
                triggered.append(watch)

        return ExecutionState(triggered_points=triggered, clock=self.get_clock_value(), cycles=0)

    def get_clock_value(self):
        """
        Looks up the value of `self.clock` in `self.value_store`
        Returns None if self.clock is None (circuit doesn't have a clock)
        """
        if self.clock is not None:
            return self.value_store.get_value(self.clock)
        return None

    def rewind(self, halfcycles):
        raise PythonSimulatorException("Reversing not currently supported")

    def cont(self):
        cycles = 0
        while True:
            self.__step()
            state = self.evaluate()
            if not state.clock:
                cycles += 1

            if state.triggered_points:
                return ExecutionState(triggered_points=state.triggered_points, clock=state.clock, cycles=cycles)

    def add_watchpoint(self, bit, scope, value=None):
        self.watchpoints.append(WatchPoint(bit, scope, self, value))
        return self.watchpoints[-1].idx

    def delete_watchpoint(self, num):
        for w in self.watchpoints:
            if w.idx == num:
                del w
                return True

        return False

    def __call__(self, *largs):
        circuit = self.main_circuit

        j = 0
        for name, port in circuit.interface.ports.items():
            if port.is_output():
                val = largs[j]
                if isinstance(port, Array):
                    n = type(port).N
                    val = BitVector[n](val)
                self.set_value(getattr(circuit, name), val)
                j += 1

        self.evaluate()

        outs = []
        for name, port in circuit.interface.ports.items():
            if port.is_input():
                val = self.get_value(getattr(circuit, name))
                val = seq2int(val) if isinstance(val, list) else int(val)
                outs.append(val)

        if len(outs) == 1:
            return outs[0]
        return tuple(outs)



import os
import types
import operator
from functools import reduce
from collections import OrderedDict
from collections.abc import Sequence
from ..compiler import Compiler
from ..ref import DefnRef, TupleRef
from ..compatibility import IntegerTypes
from ..bit import Digital
from ..clock import Clock, Enable, Reset
from ..array import Array
from ..bits import SInt
from ..tuple import Tuple
from ..is_definition import isdefinition
from magma.passes.clock import drive_undriven_clock_types_in_inst
from ..logging import root_logger
from .util import get_codegen_debug_info, make_relative
from ..config import config, EnvConfig


config._register(
    verilog_backend_log_level=EnvConfig(
        "MAGMA_VERILOG_BACKEND_LOG_LEVEL", "WARN"),
)

_logger = root_logger().getChild("verilog_backend")
_logger.setLevel(config.verilog_backend_log_level)

#__all__  = ['hstr', 'bstr']
__all__  = ['hstr']


def _verilog_name_of_ref(ref):
    if isinstance(ref, TupleRef):
        return _verilog_name_of_ref(ref.tuple.name) + "_" + str(ref.index)
    return ref.qualifiedname("_")


# return the hex character for int n
def hex(n):
    if n < 10: return chr(ord('0')+n)
    else:      return chr(ord('A')+n-10)

# return a hex string reprenting n
def hstr(n, bits):
    format = "%d'h" % bits
    nformat = []
    n &= (1 << bits)-1
    for i in range((bits+3)//4):
        nformat.append(n%16)
        n //= 16
    nformat.reverse()
    return format + reduce(operator.add, map(hex, nformat))

def bstr(n, bits):
    if bits == 1:
        return "1'b1" if init else "1'b0"
    format = "%d'b" % bits
    nformat = []
    n &= (1 << bits)-1
    for i in range(bits):
        nformat.append(n%2)
        n //= 2
    nformat.reverse()
    return format + reduce(operator.add, map(hex, nformat))

# return the verilog name of a data value
def vname(t):
    if isinstance(t, Digital):
        if t is type(t).VCC: return "1'b1"
        if t is type(t).GND: return "1'b0"
    if t.const():
        return f"{len(t)}'d{int(t)}'"

    if isinstance(t, Array):
        # print(t.ts)
        if not t.iswhole():
            # the sequence of values is concantenated
            t = [vname(i) for i in t.ts]
            t.reverse()
            return '{' + ','.join(t) + '}'

    assert not t.anon(), (t.name)
    return _verilog_name_of_ref(t.name)

# return the verilog declaration for the data type
def vdecl(t):
    if isinstance(t, Array):
        signed = "signed " if isinstance(t, SInt) else ""
        return '{}[{}:{}]'.format(signed, t.N-1, 0)
    else:
        assert isinstance(t, Digital)
        return ""

# return the verilog module args
def vmoduleargs(self):
    def append(args, port, name):
        if   port.is_input():  d = "output"
        elif port.is_output(): d = "input"
        else: d = "inout"
        args.append("%s %s %s" % (d, vdecl(port), name))

    args = []
    for name, port in self.ports.items():
        if isinstance(port, Tuple):
            for i in range(len(port)):
                append(args, port[i], vname(port[i]))
        else:
            append(args, port, name)
    return args


def compileinstance(self):

    # print('compileinstance', str(self), str(type(self)))

    def arg(k,v):
        if not isinstance(v, str): v = str(v)
        return '.%s(%s)' % (k, v)

    args = []
    debug_str = ""
    for k, v in self.interface.ports.items():
        if getattr(v, "debug_info", False) and get_codegen_debug_info():
            filename, lineno, module = (
                v.debug_info.filename, v.debug_info.lineno, v.debug_info.module
            )
        #print('arg', k, v,)
        if v.is_input():
            # find the output connected to v
            w = v.trace()
            if w is None:
                _logger.warning(f'{v.debug_name} not connected')
                continue
            v = w
        if isinstance(v, Tuple):
            for i in range(len(v)):
                args.append(arg('%s_%s' %
                    (v[i].name.tuple.name, v[i].name.index), vname(v[i])))
        elif isinstance(k, IntegerTypes):
            args.append( vname(v) )
        else:
            args.append( arg(k,vname(v)) )
        if getattr(v, "debug_info", False) and get_codegen_debug_info():
            debug_str += f"// Argument {k}({vname(v)}) wired at {make_relative(filename)}:{lineno}\n"

    params = []
    for k, v in self.kwargs.items():
       if k not in {'loc', 'name', 'T'}:
           if isinstance(v, tuple):
               v = hstr(v[0], v[1])
           params.append(arg(k, v))
    params = sorted(params)

    #s = '(* loc="%d,%d/%d" *)\n' % self.loc if self.loc else ""

    s = str(self.__class__.verilog_name)
    if len(params):
        if len(params) > 2:
            s += ' #(' + ",\n".join(params) + ')'
        else:
            s += ' #(' + ", ".join(params) + ')'
    s += ' ' + str(self.name)

    return debug_str + '%s (%s)' % (s, ', '.join(args))

def compiledefinition(cls):
    if cls.verilogFile:
       return cls.verilogFile

    # for now only allow Bit or Array(n, Bit)
    for name, port in cls.interface.ports.items():
        if isinstance(port, Array):
            if not issubclass(port.T, Digital):
                raise Exception(f'Argument {cls.__name__}.{name} of type {type(port)} is not supported, the verilog backend only supports simple 1-d array of bits of the form Array(N, Bit)')


    args = ', '.join(vmoduleargs(cls.interface))
    s = ''
    if (
        get_codegen_debug_info() and
        cls.debug_info and
        cls.debug_info.filename and
        cls.debug_info.lineno
    ):
        s += f'// Defined at {make_relative(cls.debug_info.filename)}:{cls.debug_info.lineno}\n'
    s += 'module %s (%s);\n' % (cls.verilog_name, args)
    if cls.verilog:
        s += cls.verilog + '\n'
        if cls.verilogLib:
            import re
            for libName in cls.verilogLib:
                if re.search("\\.v$",libName):
                    with open(libName,'r') as libFile:
                        s = libFile.read() + s
                else:
                    s = libName + s
    else:
        def wire(port):
            return 'wire %s %s;\n' % (vdecl(port), vname(port))

        # declare a wire for each instance output
        for instance in cls.instances:
            for port in instance.interface.ports.values():
                if isinstance(port, Tuple):
                    for i in range(len(port)):
                        s += wire(port[i])
                else:
                    if not port.is_input():
                        s += wire(port)

        #print('compile instances')
        # emit the structured verilog for each instance
        for instance in cls.instances:
            with cls.open():
                drive_undriven_clock_types_in_inst(cls, instance)
            if getattr(instance, "debug_info", False) and \
                    instance.debug_info.filename and instance.debug_info.lineno and \
                    get_codegen_debug_info():
                s += f"// Instanced at {make_relative(instance.debug_info.filename)}:{instance.debug_info.lineno}\n"
            s += compileinstance(instance) + ";\n"

        # assign to module output arguments
        for port in cls.interface.ports.values():
            if port.is_input():
                output = port.trace()
                if output is not None:
                    if isinstance(output, Tuple):
                        for name, input in cls.interface.ports.items():
                            if input.is_input():
                                output = input.trace()
                                assert isinstance(output, Tuple)
                                for i in range(len(input)):
                                    iname = vname(input[i])
                                    oname = vname(output[i])
                                    s += 'assign %s = %s;\n' % (iname, oname)
                    else:
                        iname = vname(port)
                        oname = vname(output)
                        if getattr(port, "debug_info", False) and get_codegen_debug_info():
                            s += f"// Wired at {make_relative(port.debug_info.filename)}:{port.debug_info.lineno}\n"
                        s += 'assign %s = %s;\n' % (iname, oname)
                else:
                    _logger.warning(f"{cls.__name__}.{port.name} is unwired")

    s += "endmodule\n"

    return s

def find(circuit, defn):
    if not isdefinition(circuit):
        return defn
    for i in circuit.instances:
        find(type(i), defn)
    name = circuit.verilog_name
    if name not in defn:
        defn[name] = circuit
    return defn


def compile(main):
    defn = find(main,OrderedDict())

    code = ''

    for k, v in defn.items():
         _logger.debug(f'compiling circuit {k}')
         code += compiledefinition(v) + '\n'
    return code


class VerilogCompiler(Compiler):
    def suffix(self):
        if hasattr(self.main, "verilog_file_name") and \
           os.path.splitext(self.main.verilog_file_name)[-1] == ".sv":
            return "sv"
        return "v"

    def generate_code(self):
        return compile(self.main)



import io
from typing import Dict

from magma.backend.coreir.insert_coreir_wires import insert_coreir_wires
from magma.backend.coreir.insert_wrap_casts import insert_wrap_casts
from magma.backend.mlir.compile_to_mlir import compile_to_mlir
from magma.backend.mlir.compile_to_mlir_opts import CompileToMlirOpts
from magma.backend.mlir.mlir_to_verilog import (
    MlirToVerilogOpts,
    mlir_to_verilog,
)
from magma.circuit import DefineCircuitKind
from magma.common import slice_opts
from magma.compiler import Compiler
from magma.passes.clock import wire_clocks
from magma.passes.elaborate_tuples import elaborate_tuples
from magma.passes.finalize_whens import finalize_whens
from magma.passes.raise_logs_as_exceptions import raise_logs_as_exceptions_pass


class MlirCompiler(Compiler):
    def __init__(self, main: DefineCircuitKind, basename: str, opts: Dict):
        # TODO(rsetaluri): Make this a better error.
        assert "basename" not in opts
        opts["basename"] = basename
        self._compile_to_mlir_opts = slice_opts(
            opts, CompileToMlirOpts, keep=True
        )
        self._mlir_to_verilog_opts = slice_opts(
            opts, MlirToVerilogOpts, keep=True
        )
        super().__init__(main, basename, opts)

    def suffix(self):
        if self.opts.get("output_verilog", False):
            if self.opts.get("sv", False):
                return "sv"
            return "v"
        return "mlir"

    def _run_passes(self):
        if self.opts.get("flatten_all_tuples", False):
            elaborate_tuples(self.main)

        insert_coreir_wires(self.main, flatten=False)
        wire_clocks(self.main)
        # NOTE(leonardt): finalizing whens must happen after any
        # passes that modify the circuit.  This is because passes
        # could introduce more conditional logic, or they could
        # trigger elaboration on values used in existing coditiona
        # logic (which modifies the when builder)
        finalize_whens(self.main)
        raise_logs_as_exceptions_pass(self.main)

    def compile(self):
        self._run_passes()
        with open(f"{self.basename}.mlir", "w") as f:
            compile_to_mlir(
                self.main, sout=f, opts=self._compile_to_mlir_opts)
        if not self.opts.get("output_verilog", False):
            return
        with open(f"{self.basename}.mlir", "rb") as fi:
            with open(f"{self.basename}.{self.suffix()}", "wb") as fo:
                mlir_to_verilog(fi, fo, opts=self._mlir_to_verilog_opts)



"""
Infrastructure for writing magma tests
"""
import difflib
import filecmp
import os
import pytest
import re
import shutil
import sys

from typing import Optional

from magma.config import get_debug_mode, set_debug_mode, config
from magma.conversions import bits
from magma.protocol_type import MagmaProtocolMeta, MagmaProtocol
from magma.t import Type, Direction


def check_files_equal(callee_file, file1_name, file2_name):
    """
    Check if file1 == file2 where file1 and file2 are in the same path as
    callee_file (typically __file__ for the script calling this function)
    """
    file_path = os.path.dirname(callee_file)
    file1_path = os.path.join(file_path, file1_name)
    file2_path = os.path.join(file_path, file2_name)
    print(file1_path, file2_path)
    result = filecmp.cmp(file1_path, file2_path, shallow=False)
    if not result:
        with open(file1_path, "r") as file1:
            with open(file2_path, "r") as file2:
                diff = difflib.unified_diff(
                    file2.readlines(),
                    file1.readlines(),
                    fromfile=file2_name,
                    tofile=file1_name,
                )
                for line in diff:
                    sys.stderr.write(line)
    return result


class _MagmaDebugSection:
    def __init__(self):
        self.__restore = get_debug_mode()

    def __enter__(self):
        set_debug_mode(True)

    def __exit__(self, typ, value, traceback):
        set_debug_mode(self.__restore)


def magma_debug_section():
    return _MagmaDebugSection()


def _unformat(msg):
    # Remove escaped formatting, e.g. bold or color in stdout log.
    ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    return ansi_escape.sub('', str(msg))


def has_log(caplog, level=None, msg=None):
    if level is None and msg is None:
        return len(caplog.records) > 0
    if level and not msg:
        gen = (log.levelname == level for log in caplog.records)
    elif msg and not level:
        gen = (_unformat(log.msg) == _unformat(msg) for log in caplog.records)
    else:
        gen = (log.levelname == level and _unformat(log.msg) == _unformat(msg)
               for log in caplog.records)
    return any(gen)


def has_debug(caplog, msg=None):
    return has_log(caplog, "DEBUG", msg)


def has_info(caplog, msg=None):
    return has_log(caplog, "INFO", msg)


def has_error(caplog, msg=None):
    return has_log(caplog, "ERROR", msg)


def has_warning(caplog, msg=None):
    return has_log(caplog, "WARNING", msg)


class SimpleMagmaProtocolMeta(MagmaProtocolMeta):
    _CACHE = {}

    def _to_magma_(cls):
        return cls.T

    def _qualify_magma_(cls, direction: Direction):
        return cls[cls.T.qualify(direction)]

    def _flip_magma_(cls):
        return cls[cls.T.flip()]

    def _from_magma_value_(cls, val: Type):
        return cls(val)

    def __getitem__(cls, T):
        try:
            base = cls.base
        except AttributeError:
            base = cls
        dct = {"T": T, "base": base}
        derived = type(cls)(f"{base.__name__}[{T}]", (cls,), dct)
        return SimpleMagmaProtocolMeta._CACHE.setdefault(T, derived)

    def __repr__(cls):
        return str(cls)

    def __str__(cls):
        return cls.__name__


class SimpleMagmaProtocol(MagmaProtocol, metaclass=SimpleMagmaProtocolMeta):
    def __init__(self, val: Optional[Type] = None):
        if val is None:
            val = self.T()
        self._val = val

    def _get_magma_value_(self):
        return self._val

    def non_standard_operation(self):
        v0 = self._val << 2
        v1 = bits(self._val[0], len(self.T)) << 1
        return SimpleMagmaProtocol(v0 | v1 | bits(self._val[0], len(self.T)))


def with_config(key, value):

    def fixture():
        prev_value = getattr(config, key)
        setattr(config, key, value)
        yield
        setattr(config, key, prev_value)

    return pytest.fixture(fixture)


def check_gold(callee_file, filename):
    try:
        return check_files_equal(callee_file,
                                 f"build/{filename}",
                                 f"gold/{filename}")
    except FileNotFoundError:
        return False


def update_gold(callee_file, filename):
    file_path = os.path.dirname(callee_file)
    return shutil.copy(f"{file_path}/build/{filename}",
                       f"{file_path}/gold/{filename}")



from coreir import Context


_coreir_context = Context()
_module_map = {}  # map from context to modules


def coreir_context():
    global _coreir_context
    return _coreir_context


def reset_coreir_context():
    global _coreir_context
    global _module_map
    try:
        del _module_map[_coreir_context]
    except KeyError:  # _coreir_context (singleton) not mapped
        pass
    _coreir_context.delete()
    _coreir_context = Context()


def module_map():
    global _module_map
    return _module_map



from abc import ABC, abstractmethod
from os import getenv


class ConfigBase(ABC):
    @abstractmethod
    def get(self):
        raise NotImplementedError()

    @abstractmethod
    def set(self, value):
        raise NotImplementedError()


class RuntimeConfig(ConfigBase):
    def __init__(self, init):
        self.value = init

    def get(self):
        return self.value

    def set(self, value):
        self.value = value


class EnvConfig(RuntimeConfig):
    """
    Class for capturing runtime variables which can be read from environment
    variables.

    @env_key: Key for environment variable. Can be different than magma runtime
    config key.

    @default: Default value for variable if not found in environment.

    @typ: Optional argument for the type of this variable. Since all environment
    variables are stored as strings, specifying this argument automatically
    promotes strings to the provided type.
    """
    def __init__(self, env_key, default, typ=None):
        init = getenv(env_key, default)
        if typ is not None:
            init = typ(init)
        super().__init__(init)
        self.env_key = env_key
        self.default = default
        self.typ = typ

    def reset(self, default=None):
        if not default:
            default = self.default
        value = getenv(self.env_key, default)
        if self.typ is not None:
            value = typ(value)
        self.set(value)


class ConfigManager:
    __entries = {}

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            ConfigManager.__entries[key] = value

    def _register(self, **kwargs):
        for key, value in kwargs.items():
            if key in ConfigManager.__entries:
                raise RuntimeError(f"Config with key '{key}' already exists")
            ConfigManager.__entries[key] = value

    def register(self, **kwargs):
        self._register(**kwargs)

    def __get(self, key):
        return ConfigManager.__entries[key].get()

    def __set(self, key, value):
        ConfigManager.__entries[key].set(value)

    def __getattr__(self, key):
        return self.__get(key)

    def __setattr__(self, key, value):
        self.__set(key, value)

    def __getitem__(self, key):
        return self.__get(key)

    def __setitem__(self, key, value):
        self.__set(key, value)


config = ConfigManager(
    compile_dir=RuntimeConfig("normal"),
    debug_mode=RuntimeConfig(False)
)


def get_debug_mode():
    return config.debug_mode


def get_compile_dir():
    return config.compile_dir


def set_debug_mode(value):
    config.debug_mode = value


def set_compile_dir(value):
    config.compile_dir = value



from collections import OrderedDict
import json
from hwtypes import BitVector
from magma.array import Array
from magma.bit import Digital
from magma.clock import Clock, AsyncReset, AsyncResetN, ClockTypes
from magma.ref import (ArrayRef, DefnRef, TupleRef, InstRef, NamedRef,
                       PortViewRef)
from magma.tuple import Tuple
from magma.protocol_type import magma_type, magma_value
from magma.backend.util import make_relative
from magma.t import Kind


class CoreIRBackendError(RuntimeError):
    pass


def _tuple_key_to_string(k):
    try:
        int(k)
        return f"_{k}"
    except ValueError:
        pass
    return k


def magma_name_to_coreir_select(name):
    if isinstance(name, InstRef):
        return f"{name.inst.name}.{name.name}"
    if isinstance(name, DefnRef):
        return f"self.{name.name}"
    if isinstance(name, ArrayRef):
        array_name = magma_name_to_coreir_select(name.array.name)
        if isinstance(name.index, int):
            idx_str = name.index
        else:
            assert isinstance(name.index, slice)
            idx_str = f"{name.index.start}:{name.index.stop}"
        return f"{array_name}.{idx_str}"
    if isinstance(name, TupleRef):
        tuple_name = magma_name_to_coreir_select(name.tuple.name)
        key_name = _tuple_key_to_string(name.index)
        return f"{tuple_name}.{key_name}"
    if isinstance(name, PortViewRef):
        # get select in its container definition
        inner_select = magma_name_to_coreir_select(name.view.port.name)
        return name.view.get_hierarchical_coreir_select() + inner_select
    raise NotImplementedError((name, type(name)))


def magma_port_to_coreir_port(port):
    if isinstance(port, Slice):
        return port.get_coreir_select()
    return magma_name_to_coreir_select(port.name)


def check_magma_type(type_, error_msg=""):
    type_ = magma_type(type_)
    if issubclass(type_, Array):
        msg = error_msg.format("Array({}, {})").format(str(type_.N), "{}")
        check_magma_type(type_.T, msg)
        return
    if issubclass(type_, Tuple):
        for (k, t) in zip(type_.keys(), type_.types()):
            msg = error_msg.format("Tuple({}:{})".format(k, "{}"))
            check_magma_type(t, msg)
        return
    if issubclass(type_, Digital):
        return
    raise CoreIRBackendError(error_msg.format(str(type_)))


def check_magma_interface(interface):
    # For now only allow Bit, Array, or Tuple.
    for name, port in interface.ports.items():
        check_magma_type(
            type(port), f"{name}: {type(port)} not supported by CoreIR backend"
        )


def magma_type_to_coreir_type(context, type_):
    type_ = magma_type(type_)

    if issubclass(type_, Array):
        return context.Array(type_.N,
                             magma_type_to_coreir_type(context, type_.T))
    if issubclass(type_, Tuple):
        zipped = zip(type_.keys(), type_.types())
        return context.Record({
            _tuple_key_to_string(k): magma_type_to_coreir_type(context, t)
            for (k, t) in zipped
        })
    if type_.is_input():
        if issubclass(type_, Clock):
            return context.named_types[("coreir", "clk")]
        if issubclass(type_, (AsyncReset, AsyncResetN)):
            return context.named_types[("coreir", "arst")]
        return context.Bit()
    if type_.is_output():
        if issubclass(type_, Clock):
            return context.named_types[("coreir", "clkIn")]
        if issubclass(type_, (AsyncReset, AsyncResetN)):
            return context.named_types[("coreir", "arstIn")]
        return context.BitIn()
    return context.BitInOut()


def magma_interface_to_coreir_module_type(context, interface):
    args = OrderedDict()
    for name, port in interface.ports.items():
        args[name] = magma_type_to_coreir_type(context, type(port))
    return context.Record(args)


def python_to_coreir_param_type(context, typ):
    if typ is int:
        return context.Int()
    if typ is bool:
        return context.Bool()
    if typ is str:
        return context.String()
    if typ is BitVector:
        return context.BitVector()
    raise NotImplementedError(typ)


def make_cparams(context, params):
    cparams = {
        k: python_to_coreir_param_type(context, t)
        for k, t in params.items()
    }
    return context.newParams(cparams)


def attach_debug_info(coreir_obj, debug_info, a=None, b=None):
    def fn(k, v):
        if a and b:
            return coreir_obj.add_metadata(a, b, k, v)
        return coreir_obj.add_metadata(k, v)

    fn("filename", json.dumps(make_relative(debug_info.filename)))
    fn("lineno", json.dumps(str(debug_info.lineno)))


def map_genarg(context, value):
    if isinstance(value, type) and issubclass(value, Clock):
        if value.is_input():
            return context.named_types[("coreir", "clkIn")]
        return context.named_types[("coreir", "clk")]
    if isinstance(value, type) and issubclass(value, (AsyncReset, AsyncResetN)):
        if value.is_input():
            return context.named_types[("coreir", "arstIn")]
        return context.named_types[("coreir", "arst")]
    return value


def get_module_of_inst(context, inst, lib):
    wrapped = getattr(inst, "wrappedModule", None)
    if wrapped and wrapped.context is context:
        return wrapped
    return lib.modules[type(inst).coreir_name]


def get_inst_args(inst):
    args = {}
    for name, value in inst.kwargs.items():
        if name == "name" or name == "loc":
            continue
        if isinstance(value, tuple):
            args[name] = BitVector[value[1]](value[0])
        else:
            args[name] = value
    return args


def constant_to_value(constant):
    assert constant.const()
    if isinstance(constant, Digital):
        return int(constant)
    if isinstance(constant, Array):
        values = [constant_to_value(c) for c in constant]
        return BitVector[len(constant)](values)
    raise NotImplementedError(constant)


class Slice:
    def __init__(self, value, low, high):
        self.value = value
        self.low = low
        self.high = high

    def get_coreir_select(self,):
        return (magma_name_to_coreir_select(self.value.name) +
                f".{self.low}:{self.high}")



from hwtypes import BitVector, Bit

from .compatibility import IntegerTypes
from .wire_container import Wire  # TODO(rsetaluri): only here for b.c.
from .debug import debug_wire
from .logging import root_logger
from .protocol_type import magma_value

from magma.wire_container import WiringLog


_logger = root_logger()


_CONSTANTS = (IntegerTypes, BitVector, Bit)


@debug_wire
def wire(o, i, debug_info=None):
    o = magma_value(o)
    i = magma_value(i)

    # Circular import
    from .conversions import tuple_
    if isinstance(o, tuple):
        o = tuple_(o)
    if isinstance(i, tuple):
        i = tuple_(i)

    # Wire(o, Circuit).
    if hasattr(i, 'interface'):
        i.wire(o, debug_info)
        return

    # Replace output Circuit with its output (should only be 1 output).
    if hasattr(o, 'interface'):
        outputs = o.interface.outputs()
        if len(outputs) != 1:
            _logger.error(
                WiringLog(f"Can only wire circuits with one output; circuit "
                          f"`{{}}` has outputs "
                          f"{[output.name.name for output in outputs]}", o),
                debug_info=debug_info
            )
            return
        o = outputs[0]

    # If o is an input.
    if not isinstance(o, _CONSTANTS) and o.is_input():
        # If i is not an input.
        if isinstance(i, _CONSTANTS) or not i.is_input():
            # Flip i and o.
            i, o = o, i

    i_T, o_T = type(i), type(o)
    if not i_T.is_wireable(o_T):
        _logger.error(
            WiringLog(f"Cannot wire {{}} ({o_T}) to {{}} ({i_T})",
                      o, i),
            debug_info=debug_info
        )
        return

    # Wire(o, Type).
    i.wire(o, debug_info)


@debug_wire
def unwire(input, output=None, debug_info=None):
    input.unwire(output, debug_info)
